#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from hwt.hdl.types.defs import INT, STR, BOOL, FLOAT64
from hwt.hdl.value import HValue


class Param():
    """
    Class used to mark object as a configuration of HDL module. (
    The parameter instance will not appear on :class:`hwt.synthesizer.unit.Unit` instance,
    instead the value will appear.
    The parameter instance will be stored
    in ._params property of Unit/Interface object)

    :ivar ~._initval: value of the parameter which should be used for intialization
    :attention: the actual value is then store on parent object instance
    :ivar ~._name: name of parameter on parent Unit/Interface instance
    :ivar ~._parent: parent object instance
    """
    __slots__ = ["_initval", "_name", "hdl_name", "_parent"]

    def __init__(self, initval):
        self._initval = initval
        self._name = None
        self.hdl_name = None
        self._parent = None

    def get_hdl_type(self):
        v = self.get_value()
        INT32_MAX = 2 ** (32 - 1) - 1
        INT32_MIN = -2 ** (32 - 1)

        if isinstance(v, HValue):
            return v._dtype
        elif isinstance(v, bool):
            return BOOL
        elif isinstance(v, str):
            return STR
        elif isinstance(v, int) and v >= INT32_MIN and v <= INT32_MAX:
            return INT
        elif isinstance(v, float):
            return FLOAT64
        else:
            return None

    def get_hdl_value(self):
        t = self.get_hdl_type()
        v = self.get_value()
        if t is None:
            t = STR
            v = t.from_py(str(v))
        else:
            if not isinstance(v, HValue):
                v = t.from_py(v)
        return v

    def get_value(self):
        return getattr(self._parent, self._name)

    def set_value(self, v):
        setattr(self._parent, self._name, v)

    def __repr__(self):
        return "<%s at 0x%x %s=%s>" % (
            self.__class__.__name__,
            id(self),
            "<unspecified name>" if self._name is None else self._name,
            repr(self.get_value()))



from copy import copy
from typing import Optional, List, Dict, Tuple, Set

from hdlConvertorAst.hdlAst._structural import HdlCompInst
from hwt.doc_markers import internal
from hwt.hdl.portItem import HdlPortItem
from hwt.synthesizer.dummyPlatform import DummyPlatform
from hwt.synthesizer.exceptions import IntfLvlConfErr
from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase
from hwt.synthesizer.interfaceLevel.propDeclrCollector import PropDeclrCollector
from hwt.synthesizer.interfaceLevel.unitImplHelpers import UnitImplHelpers, \
    _default_param_updater
from hwt.synthesizer.rtlLevel.netlist import RtlNetlist
from ipCorePackager.constants import DIRECTION


class HdlConstraintList(list):
    """
    Containers of hw design constraints
    """
    pass


class Unit(PropDeclrCollector, UnitImplHelpers):
    """
    Container of the netlist with interfaces
    and internal hierarchical structure

    :cvar ~._serializeDecision: function to decide if HDL object derived from
        this unit should be serialized or not, if None all is always serialized
    :cvar ~._PROTECTED_NAMES: set of names which can not be overridden
    :ivar ~._interfaces: all public interfaces
    :type ~._interfaces: List[Interface]
    :ivar ~._private_interfaces: all internal interfaces
        which are not accessible from outside of unit
    :type _private_interfaces: List[Interface]
    :ivar ~._units: all units defined on this object
    :type ~._units: List[Unit]
    :ivar ~._params: all params defined on this object
    :type ~._params: List[Param]
    :ivar ~._constraints: additional HW specifications
    :ivar ~._parent: parent object
    :type ~._parent: Optional[Unit]
    :ivar ~._lazy_loaded: container of RTL object which were lazy loaded
        in implementation phase (this object has to be returned
        from :func:`~._to_rtl` of parent before it it's own objects)
    :ivar ~._shared_component_with: Optional tuple of the other :class:`hwt.synthesizer.unit.Unit` instance
        which produces an exactly same component in HDL and interface
        signal map current to shared and shared to current
    :type ~._shared_component_with: Optional[Tuple[Unit,
        Dict[Interface, Interface],
        Dict[Interface, Interface]]]
    :attention: if :func:`~._shared_component_with` is not None the body
        of this instance is not generated at all
        and the component from :func:`~._shared_component_with` is used instead
    :ivar ~._target_platform: meta-informations about target platform
    :ivar ~._name: a name of this component
    :ivar ~._hdl_module_name: a name of HDL module for this component
        (vhdl entity name, Verilog module name)
    """

    _serializeDecision = None
    # properties which are used internally by this library
    _PROTECTED_NAMES = set([
        "_PROTECTED_NAMES",
        "_name", "_hdl_module_name",
        "_interfaces", "_private_interfaces",
        "_units", "_params", "_parent", "_constraints",
        "_lazy_loaded", "_ctx", "_shared_component_with",
        "_target_platform", "_store_manager",
    ])

    def __init__(self, hdl_name_override:Optional[str]=None):
        self._parent = None
        self._name = None
        self._shared_component_with = None
        self._hdl_module_name = None
        assert hdl_name_override is None or isinstance(hdl_name_override, str), hdl_name_override
        self._hdl_name_override = hdl_name_override
        self._lazy_loaded = []
        self._ctx = RtlNetlist(self)
        self._constraints = HdlConstraintList()
        self._loadConfig()

    @internal
    def _loadInterface(self, i, isExtern):
        i._loadDeclarations()
        i._setAsExtern(isExtern)

    @internal
    def _loadDeclarations(self):
        """
        Load all declarations from _decl() method, recursively
        for all interfaces/units.
        """
        if not hasattr(self, "_interfaces"):
            self._interfaces = []
        if not hasattr(self, "_private_interfaces"):
            self._private_interfaces = []
        if not hasattr(self, "_units"):
            self._units = []
        self._setAttrListener = self._declrCollector
        self._declr()
        self._setAttrListener = None
        for i in self._interfaces:
            self._loadInterface(i, True)

        # if I am a unit load subunits
        for u in self._units:
            u._loadDeclarations()

    @internal
    def _registerIntfInImpl(self, iName, intf):
        """
        Register interface in implementation phase
        """
        self._registerInterface(iName, intf, isPrivate=True)
        self._loadInterface(intf, False)
        intf._signalsForInterface(
            self._ctx, None, self._store_manager.name_scope)

    def _getDefaultName(self) -> str:
        return self.__class__.__name__

    def _get_hdl_doc(self) -> Optional[str]:
        if self.__doc__ is not Unit.__doc__:
            return self.__doc__

    @internal
    def _to_rtl(self, target_platform: DummyPlatform,
                store_manager: "StoreManager"):
        """
        synthesize all subunits, make connections between them,
        build verilog like module/vhdl like entity and architecture for this unit
        """
        if self._hdl_module_name is None:
            if self._hdl_name_override:
                self._hdl_module_name = self._hdl_name_override
            else:
                self._hdl_module_name = self._getDefaultName()
        if self._name is None:
            self._name = self._getDefaultName()
        self._target_platform = target_platform
        self._store_manager = store_manager
        do_serialize_this, replacement = store_manager.filter.do_serialize(
            self)
        if replacement is not None:
            assert not do_serialize_this
            assert len(self._interfaces) == len(replacement._interfaces), \
                "No lazy loaded interfaces declared in _impl()"
            copy_HdlModuleDec(replacement, self)
            yield False, self
            self._cleanAsSubunit()
            self._units = None
            self._private_interfaces = None
            intf_map_repl_to_self = shared_comp_build_interface_map(
                replacement, self)
            intf_map_self_to_repl = {
                v: k
                for k, v in intf_map_repl_to_self.items()}
            self._shared_component_with = replacement, \
                intf_map_self_to_repl, intf_map_repl_to_self
            return

        for proc in target_platform.beforeToRtl:
            proc(self)

        mdec = self._ctx.create_HdlModuleDec(
            self._hdl_module_name, store_manager, self._params)
        mdec.origin = self
        mdec.doc = self._get_hdl_doc()

        # prepare signals for interfaces
        for i in self._interfaces:
            if i._isExtern:
                ei = self._ctx.interfaces
            else:
                ei = None
            # we are reversing direction because we are looking
            # at the interface from inside of component
            i._signalsForInterface(
                self._ctx, ei,
                store_manager.name_scope, reverse_dir=True)
        store_manager.hierarchy_pop(mdec)

        if do_serialize_this:
            # prepare subunits
            for u in self._units:
                yield from u._to_rtl(target_platform, store_manager)

            # now every sub unit has a HdlModuleDec prepared
            for u in self._units:
                subUnitName = u._name
                u._signalsForSubUnitEntity(self._ctx, "sig_" + subUnitName)

            for proc in target_platform.beforeToRtlImpl:
                proc(self)

        try:
            store_manager.hierarchy_push(mdec)
            if do_serialize_this:
                self._loadImpl()
                yield from self._lazy_loaded

                if not self._ctx.interfaces:
                    raise IntfLvlConfErr(
                        "Can not find any external interface for unit %s"
                        "- unit without interfaces are not synthetisable"
                        % self._name)

            for proc in target_platform.afterToRtlImpl:
                proc(self)

            mdec.params.sort(key=lambda x: x.name)
            mdec.ports.sort(key=lambda x: x.name)
            if do_serialize_this:
                # synthesize signal level context
                mdef = self._ctx.create_HdlModuleDef(
                    target_platform, store_manager)
                mdef.origin = self

            for intf in self._interfaces:
                if intf._isExtern:
                    # reverse because other components
                    # looks at this interface from the outside
                    intf._reverseDirection()

            if do_serialize_this:
                store_manager.write(mdef)

            yield True, self

            # after synthesis clean up interface so this :class:`hwt.synthesizer.unit.Unit` object can be
            # used elsewhere
            self._cleanAsSubunit()
            if do_serialize_this:
                self._checkCompInstances()

            for proc in target_platform.afterToRtl:
                proc(self)
        finally:
            store_manager.hierarchy_pop(mdec)

    def _updateParamsFrom(self, otherObj: PropDeclrCollector,
                          updater=_default_param_updater,
                          exclude: Optional[Tuple[Set[str], Set[str]]]=None,
                          prefix=""):
        """
        :note: doc in
            :func:`hwt.synthesizer.interfaceLevel.propDeclCollector._updateParamsFrom`
        """
        return PropDeclrCollector._updateParamsFrom(self, otherObj,
                                                    updater, exclude, prefix)

    @internal
    def _checkCompInstances(self):
        cInstances = [o for o in self._ctx.arch.objs
                      if isinstance(o, HdlCompInst)]
        cInst_cnt = len(cInstances)
        unit_cnt = len(self._units)
        if cInst_cnt != unit_cnt:
            # resolve the error message
            inRtl = set(x.name for x in cInstances)
            inIntf = set(x._name for x in self._units)
            cls_name = self.__class__.__name__
            if cInst_cnt > unit_cnt:
                diff = inRtl - inIntf
                raise IntfLvlConfErr(
                    f"{cls_name:s}, {self._name:s}: unit(s) were found in HDL but were"
                    f" not registered {diff}")
            else:
                assert cInst_cnt < unit_cnt
                diff = inIntf - inRtl
                raise IntfLvlConfErr(
                    f"{cls_name:s}, {self._name:s}: _to_rtl: unit(s) are missing in produced HDL {diff}")


def copy_HdlModuleDec_interface(orig_i: InterfaceBase, new_i: InterfaceBase,
                                ports: List[HdlPortItem], new_u: Unit):
    new_i._direction = orig_i._direction
    if orig_i._hdl_port is not None:
        s = orig_i._sigInside
        assert s is not None, (
            "the component which shares a body with this component"
            " is actually some parent of this component")
        pi = copy(orig_i._hdl_port)
        pi.unit = new_u
        ports.append(pi)
        d = pi.direction
        if d == DIRECTION.OUT:
            pi.dst = None
        elif d == DIRECTION.IN:
            pi.src = None
        else:
            raise NotImplementedError(d)
        new_i._hdl_port = pi
        new_i._sigInside = s
    else:
        for i in orig_i._interfaces:
            n_i = getattr(new_i, i._name)
            copy_HdlModuleDec_interface(i, n_i, ports, new_u)


def copy_HdlModuleDec(orig_u: Unit, new_u: Unit):
    assert not new_u._ctx.statements
    assert not new_u._ctx.interfaces
    assert not new_u._ctx.signals
    assert new_u._ctx.ent is None

    new_u._hdl_module_name = orig_u._hdl_module_name
    e = new_u._ctx.ent = copy(orig_u._ctx.ent)

    params = []
    param_by_name = {p._name: p for p in new_u._params}
    for p in e.params:
        new_p_def = copy(p)
        old_p = new_p_def.origin = param_by_name[p.origin._name]
        old_p.hdl_name = p.origin.hdl_name
        new_p_def.value = old_p.get_hdl_value()
        params.append(new_p_def)
    e.params = params

    e.ports = []
    for oi, ni in zip(orig_u._interfaces, new_u._interfaces):
        if oi._isExtern:
            copy_HdlModuleDec_interface(oi, ni, e.ports, new_u)

    # params should be already sorted
    # e.params.sort(key=lambda x: x.name)
    e.ports.sort(key=lambda x: x.name)


def _shared_comp_build_interface_map_list(replacement: List[InterfaceBase],
                                          substituted: List[InterfaceBase],
                                          res: Dict[InterfaceBase, InterfaceBase]):
    assert len(replacement) == len(substituted)
    for r, s in zip(replacement, substituted):
        assert r._name == s._name, (r._name, s._name)
        res[r] = s
        if r._interfaces:
            _shared_comp_build_interface_map_list(
                r._interfaces, s._interfaces, res)


def shared_comp_build_interface_map(replacement_u: Unit, substituted_u: Unit):
    """
    Build a dictionary which maps
    interface of replacement_u to interface of substituted_u
    """
    res = {}
    _shared_comp_build_interface_map_list(
        replacement_u._interfaces, substituted_u._interfaces, res)
    return res



from hdlConvertorAst.hdlAst._expr import HdlValueId, HdlValueInt
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model.utils import hdl_call, \
    hdl_map_asoc, hdl_index
from hwt.hdl.types.array import HArray
from hwt.hdl.types.bits import BITS_DEFAUTL_SIGNED, BITS_DEFAUTL_FORCEVECTOR, \
    BITS_DEFAUTL_NEGATED, Bits
from hwt.hdl.types.defs import BOOL, INT
from hwt.hdl.types.enum import HEnum
from hwt.hdl.types.hdlType import MethodNotOverloaded
from hwt.hdl.types.float import HFloat


class ToHdlAstHwt_types():
    """
    part of ToHdlAstSimModel responsible for type serialization
    """
    BOOL = HdlValueId("BOOL", obj=BOOL)
    INT = HdlValueId("INT", obj=INT)
    BITS = HdlValueId("Bits", obj=Bits)

    def does_type_requires_extra_def(self, t, other_types):
        try:
            return t._as_hdl_requires_def(self, other_types)
        except MethodNotOverloaded:
            pass
        return isinstance(t, HEnum) and t not in other_types

    def as_hdl_HdlType_array(self, typ: HArray, declaration=False):
        assert not declaration, "declaration should not be required"
        t = self.as_hdl_HdlType(typ.element_t, declaration=declaration)
        return hdl_index(t, HdlValueInt(int(typ.size), None, None))

    def as_hdl_HdlType_bits(self, typ: Bits, declaration=False):
        if declaration:
            raise NotImplementedError()
        elif typ == BOOL:
            return self.BOOL
        elif typ == INT:
            return self.INT

        w = typ.bit_length()
        assert isinstance(w, int), w

        def add_kw(name, val):
            kw = hdl_map_asoc(HdlValueId(name),
                              HdlValueInt(val, None, None))
            args.append(kw)

        args = [HdlValueInt(w, None, None)]
        if typ.signed is not BITS_DEFAUTL_SIGNED:
            add_kw("signed", typ.signed)
        if typ.force_vector is not BITS_DEFAUTL_FORCEVECTOR and w <= 1:
            add_kw("force_vector", typ.force_vector)
        if typ.negated is not BITS_DEFAUTL_NEGATED:
            add_kw("negated", typ.negated)

        return hdl_call(self.BITS, args)

    def as_hdl_HdlType_float(self, typ: HFloat, declaration=False):
        return hdl_call(HdlValueId("HFloat"), typ.exponent_w, typ.mantisa_w)



from operator import and_, or_, xor, add

from hwt.code_utils import _mkOp, _intfToSig
from hwt.hdl.operatorDefs import concatFn
from hwt.hdl.statements.codeBlockContainer import HdlStmCodeBlockContainer
from hwt.hdl.statements.ifContainter import IfContainer
from hwt.hdl.statements.statement import HwtSyntaxError
from hwt.hdl.statements.switchContainer import SwitchContainer
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.enum import HEnum
from hwt.hdl.types.typeCast import toHVal
from hwt.hdl.value import HValue
from hwt.math import log2ceil
from hwt.pyUtils.arrayQuery import arr_any
from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.synthesizer.rtlLevel.signalUtils.walkers import \
    discoverEventDependency


class CodeBlock(HdlStmCodeBlockContainer):
    """
    Cointainer for list of statements
    """

    def __init__(self, *statements):
        super(CodeBlock, self).__init__()
        self._register_stements(statements, self.statements)
        self.rank = sum(map(lambda s: s.rank, statements))

        if self._outputs:
            ctx = self._get_rtl_context()
            ctx.statements.add(self)


class If(IfContainer):
    """
    If statement generator
    """

    def __init__(self, cond, *statements):
        """
        :param cond: condition in if statement
        :param statements: list of statements which should be active
            if condition is met
        """
        cond_sig = _intfToSig(cond)
        if not isinstance(cond_sig, RtlSignalBase):
            raise HwtSyntaxError("Condition is not signal, it is not certain"
                                 " if this is an error or desire ", cond_sig)

        super(If, self).__init__(cond_sig)
        self.rank = 1
        self._inputs.append(cond_sig)
        cond_sig.endpoints.append(self)

        ev_dep = arr_any(discoverEventDependency(cond_sig), lambda x: True)
        self._event_dependent_from_branch = 0 if ev_dep else None

        self._register_stements(statements, self.ifTrue)
        self._get_rtl_context().statements.add(self)

    def Elif(self, cond, *statements):
        assert self.parentStm is None
        self.rank += 1
        cond_sig = _intfToSig(cond)

        ev_dep = arr_any(discoverEventDependency(cond_sig), lambda x: True)
        self._event_dependent_from_branch = len(self.elIfs) + 1 if ev_dep else None

        self._inputs.append(cond_sig)
        cond_sig.endpoints.append(self)

        case = []
        self.elIfs.append((cond_sig, case))
        self._register_stements(statements, case)

        return self

    def Else(self, *statements):
        assert self.parentStm is None
        if self.ifFalse is not None:
            raise HwtSyntaxError(
                "Else on this if-then-else statement was already used")

        self.rank += 1

        self.ifFalse = []
        self._register_stements(statements, self.ifFalse)
        return self


class Switch(SwitchContainer):
    """
    Switch statement generator
    """

    def __init__(self, switchOn):
        switchOn = _intfToSig(switchOn)
        if not isinstance(switchOn, RtlSignalBase):
            raise HwtSyntaxError("Select is not signal, it is not certain"
                                 " if this is an error or desire")
        if arr_any(discoverEventDependency(switchOn), lambda x: True):
            raise HwtSyntaxError("Can not switch on result of event operator")

        super(Switch, self).__init__(switchOn, [])
        switchOn.ctx.statements.add(self)
        self._inputs.append(switchOn)
        switchOn.endpoints.append(self)

    def add_cases(self, tupesValStms):
        """
        Add multiple case statements from iterable of tuples
        (caseVal, statements)
        """
        s = self
        for val, statements in tupesValStms:
            s = s.Case(val, statements)
        return s

    def Case(self, caseVal, *statements):
        "c-like case of switch statement"
        assert self.parentStm is None
        caseVal = toHVal(caseVal, self.switchOn._dtype)

        assert isinstance(caseVal, HValue), caseVal
        assert caseVal._is_full_valid(), "Cmp with invalid value"
        assert caseVal not in self._case_value_index, (
            "Switch statement already has case for value ", caseVal)

        self.rank += 1
        case = []
        self._case_value_index[caseVal] = len(self.cases)
        self.cases.append((caseVal, case))
        self._register_stements(statements, case)

        return self

    def Default(self, *statements):
        """c-like default of switch statement
        """
        assert self.parentStm is None
        self.rank += 1
        self.default = []
        self._register_stements(statements, self.default)
        return self


def SwitchLogic(cases, default=None):
    """
    Generate if tree for cases like (syntax sugar for large generated elifs)

    ..code-block:: python
        if cond0:
            statements0
        elif cond1:
            statements1
        else:
            default

    :param case: iterable of tuples (condition, statements)
    :param default: default statements
    """
    assigTop = None
    hasElse = False
    for cond, statements in cases:
        if isinstance(cond, (RtlSignalBase, InterfaceBase)):
            if assigTop is None:
                assigTop = If(cond,
                             statements
                           )
            else:
                assigTop = assigTop.Elif(cond, statements)
        else:
            if cond:
                if assigTop is None:
                    assigTop = statements
                else:
                    assigTop.Else(statements)
                    hasElse = True
            else:
                pass

    if assigTop is None:
        if default is None:
            return []
        else:
            return default
    else:
        if hasElse:
            return assigTop
        elif default is not None:
            assigTop = assigTop.Else(default)

        return assigTop


def In(sigOrVal, iterable):
    """
    HDL convertible "in" operator, check if any of items
    in "iterable" equals "sigOrVal"
    """
    res = None
    for i in iterable:
        i = toHVal(i)
        if res is None:
            res = sigOrVal._eq(i)
        else:
            res = res | sigOrVal._eq(i)

    assert res is not None, "argument iterable is empty"
    return res


def StaticForEach(parentUnit, items, bodyFn, name=""):
    """
    Generate for loop for static items

    :param parentUnit: unit where this code should be instantiated
    :param items: items which this "for" iterating on
    :param bodyFn: function which fn(item, index) or fn(item)
        returns (statementList, ack).
        It's content is performed in every iteration.
        When ack is high loop will fall to next iteration
    """

    items = list(items)
    itemsCnt = len(items)
    if itemsCnt == 0:
        # if there are no items there is nothing to generate
        return []
    elif itemsCnt == 1:
        # if there is only one item do not generate counter logic generate
        return bodyFn(items[0], 0)
    else:
        # if there is multiple items we have to generate counter logic
        index = parentUnit._reg(name + "for_index",
                                Bits(log2ceil(itemsCnt + 1), signed=False),
                                def_val=0)
        ackSig = parentUnit._sig(name + "for_ack")

        statementLists = []
        for i, (statementList, ack) in [(i, bodyFn(item, i))
                                        for i, item in enumerate(items)]:
            statementLists.append(statementList + [(ackSig(ack)), ])

        If(ackSig,
           If(index._eq(itemsCnt - 1),
              index(0)
              ).Else(
               index(index + 1)
           )
           )

        return Switch(index)\
            .add_cases(
            enumerate(statementLists)
        ).Default(
            bodyFn(items[0], 0)[0],
            ackSig(True)
        )


class FsmBuilder(Switch):
    """
    A syntax sugar which automatically construct the state transition switch and state register

    :ivar ~.stateReg: register with state
    """

    def __init__(self, parent, stateT, stateRegName="st"):
        """
        :param parent: parent unit where fsm should be builded
        :param stateT: enum type of state
        :param stateRegName: name of register where sate is stored
        """
        if isinstance(stateT, HEnum):
            beginVal = stateT.from_py(stateT._allValues[0])
        else:
            beginVal = 0

        self.stateReg = parent._reg(stateRegName, stateT, beginVal)
        Switch.__init__(self, self.stateReg)

    def Trans(self, stateFrom, *condAndNextState):
        """
        :param stateFrom: apply when FSM is in this state
        :param condAndNextState: tuples (condition, newState),
            last does not to have condition

        :attention: transitions has priority, first has the biggest
        :attention: if stateFrom is None it is evaluated as default
        """
        top = []
        last = True

        for cAndS in reversed(condAndNextState):
            if last is True:
                last = False
                # if this is last trans. it does not have to condition
                try:
                    condition, newvalue = cAndS
                except TypeError:
                    top = self.stateReg(cAndS)
                    continue
                top = []

            else:
                condition, newvalue = cAndS

            # building decision tree
            top = \
                If(condition,
                   self.stateReg(newvalue)
                ).Else(
                    top
                )
        if stateFrom is None:
            return Switch.Default(self, top)
        else:
            return Switch.Case(self, stateFrom, top)

    def Default(self, *condAndNextState):
        d = self.Trans(None, *condAndNextState)
        d.stateReg = self.stateReg
        return d


# variadic operator functions
And = _mkOp(and_)
Add = _mkOp(add)
Or = _mkOp(or_)
Xor = _mkOp(xor)
Concat = _mkOp(concatFn)


def ror(sig, howMany) -> RtlSignalBase:
    "Rotate right"
    if sig._dtype.bit_length() == 1:
        return sig
    return sig[howMany:]._concat(sig[:howMany])


def rol(sig, howMany) -> RtlSignalBase:
    "Rotate left"
    width = sig._dtype.bit_length()
    if width == 1:
        return sig
    return sig[(width - howMany):]._concat(sig[:(width - howMany)])


def replicate(n, v):
    return Concat(*(v for _ in range(n)))



from hwt.doc_markers import internal
from hwt.hdl.statements.assignmentContainer import HdlAssignmentContainer
from hwt.hdl.statements.codeBlockContainer import HdlStmCodeBlockContainer
from hwt.hdl.statements.statement import HdlStatement
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


@internal
def getMaxStmIdForStm(stm):
    """
    Get maximum _instId from all assignments in statement,
    used for sorting of processes in architecture
    """
    maxId = 0
    if isinstance(stm, HdlAssignmentContainer):
        return stm._instId
    else:
        for _stm in stm._iter_stms():
            maxId = max(maxId, getMaxStmIdForStm(_stm))
        return maxId


def RtlSignal_sort_key(s: RtlSignalBase):
    return (s.name, s._instId)


def HdlStatement_sort_key(stm: HdlStatement):
    if isinstance(stm, HdlStmCodeBlockContainer) and stm.name is not None:
        return (stm.name, getMaxStmIdForStm(stm))
    else:
        return ("", getMaxStmIdForStm(stm))



from itertools import zip_longest
from math import ceil, floor, inf
from typing import Union, Generator, List

from hwt.doc_markers import internal
from hwt.hdl.frameTmplUtils import TransTmplWordIterator, \
    ChoicesOfFrameParts
from hwt.hdl.transPart import TransPart
from hwt.hdl.types.array import HArray
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.hdlType import HdlType
from hwt.hdl.types.struct import HStruct
from hwt.pyUtils.arrayQuery import flatten
from pyMathBitPrecise.bit_utils import mask, get_bit_range, set_bit_range


class FrameTmpl(object):
    """
    Frame template is metainfomation about data structure, it's the template
    of transaction for specific interface

    Usuall flow of frame generatig consists of these steps:

    1. Describe format of data by HDL type (HStruct, HUnion ...)

    2. Convert it to TransTmpl to resolve addresses of each field in structure

    3. Split parts of TransTmpl into words

    4. Use parts in words to assembly frames

    :ivar ~._fieldToTPart: dictionary {HStructField: TransPart} to resolve
        this association, None by default, builded when packData
        is called and is not builded
    :note: others ivars described in __init__
    """

    def __init__(self, origin: HdlType,
                 wordWidth: int,
                 startBitAddr: int,
                 endBitAddr: int,
                 transParts: List[TransPart]):
        """
        :param origin: instance of HType (usually HStruct)
            from which this FrameTmpl was generated from
        :param wordWidth: width of word on interface
            where this template should be used
        :param startBitAddr: bit offset where this frame starts
        :param endBitAddr: bit offset where this frame ends
            (bit index of first bit behind this frame)
        :param transParts: instances of TransPart which are parts of this frame
        """
        self.origin = origin
        self.wordWidth = wordWidth
        assert startBitAddr <= endBitAddr
        self.startBitAddr = startBitAddr
        self.endBitAddr = endBitAddr
        self.parts = transParts

        self._fieldToTPart = None

        for p in self.parts:
            p.parent = self
            assert p.startOfPart >= startBitAddr, (p, startBitAddr)
            assert p.endOfPart <= endBitAddr, (p, endBitAddr)

    @staticmethod
    def framesFromTransTmpl(transaction: 'TransTmpl',
                            wordWidth: int,
                            maxFrameLen: Union[int, float]=inf,
                            maxPaddingWords: Union[int, float]=inf,
                            trimPaddingWordsOnStart: bool=False,
                            trimPaddingWordsOnEnd: bool=False) -> Generator[
                                'FrameTmpl', None, None]:
        """
        Convert transaction template into FrameTmpls

        :param transaction: transaction template used which are FrameTmpls
            created from
        :param wordWidth: width of data signal in target interface
            where frames will be used
        :param maxFrameLen: maximum length of frame in bits,
            if exceeded another frame will be created
        :param maxPaddingWords: maximum of continual padding words in frame,
            if exceed frame is split and words are cut of
        :attention: if maxPaddingWords<inf trimPaddingWordsOnEnd
            or trimPaddingWordsOnStart has to be True
            to decide where padding should be trimmed
        :param trimPaddingWordsOnStart: trim padding from start of frame
            at word granularity
        :param trimPaddingWordsOnEnd: trim padding from end of frame
            at word granularity
        """
        isFirstInFrame = True
        partsPending = False

        startOfThisFrame = 0
        assert maxFrameLen > 0
        assert maxPaddingWords >= 0
        if maxPaddingWords < inf:
            assert trimPaddingWordsOnStart or trimPaddingWordsOnEnd, \
                "Padding has to be cut off somewhere"

        it = TransTmplWordIterator(wordWidth)
        lastWordI = 0
        endOfThisFrame = maxFrameLen
        parts = []
        for wordI, word in it.groupByWordIndex(transaction, 0):
            if wordI * wordWidth >= endOfThisFrame:
                # now in first+ word behind the frame
                # cut off padding at end of frame
                paddingWords = wordI - lastWordI
                if trimPaddingWordsOnEnd and paddingWords > maxPaddingWords:
                    # cut off padding and align end of frame to word
                    _endOfThisFrame = (lastWordI + 1) * wordWidth
                else:
                    _endOfThisFrame = wordI * wordWidth

                yield FrameTmpl(transaction,
                                wordWidth,
                                startOfThisFrame,
                                _endOfThisFrame,
                                parts)

                # prepare for start of new frame
                parts = []
                isFirstInFrame = True
                partsPending = False
                # start on new word
                startOfThisFrame = _endOfThisFrame
                endOfThisFrame = startOfThisFrame + maxFrameLen
                lastWordI = wordI

            # check if padding at potential end of frame can be cut off
            if (not isFirstInFrame
                    and trimPaddingWordsOnEnd
                    and wordI - lastWordI > maxPaddingWords + 1):
                # there is too much continual padding,
                # cut it out and start new frame
                _endOfThisFrame = (lastWordI + 1) * wordWidth
                yield FrameTmpl(transaction,
                                wordWidth,
                                startOfThisFrame,
                                _endOfThisFrame,
                                parts)

                # prepare for start of new frame
                parts = []
                isFirstInFrame = True
                partsPending = False
                # start on new word
                lastWordI = wordI - 1
                startOfThisFrame = lastWordI * wordWidth
                endOfThisFrame = startOfThisFrame + maxFrameLen

            if isFirstInFrame:
                partsPending = True
                isFirstInFrame = False
                # cut off padding at start of frame
                paddingWords = wordI - lastWordI
                if trimPaddingWordsOnStart and paddingWords > maxPaddingWords:
                    startOfThisFrame += paddingWords * wordWidth

                endOfThisFrame = startOfThisFrame + maxFrameLen

            # resolve end of this part
            parts.extend(word)
            lastWordI = wordI

        # reminder in "parts" after last iteration
        endOfThisFrame = transaction.bitAddrEnd
        withPadding = not (trimPaddingWordsOnEnd or trimPaddingWordsOnStart)
        if partsPending or (withPadding
                            and endOfThisFrame != startOfThisFrame):
            # cut off padding at end of frame
            endOfLastWord = (lastWordI + 1) * wordWidth
            if endOfThisFrame < endOfLastWord:
                endOfThisFrame = endOfLastWord
            else:
                paddingWords = it.fullWordCnt(endOfLastWord, endOfThisFrame)
                if trimPaddingWordsOnEnd and paddingWords > maxPaddingWords:
                    endOfThisFrame -= paddingWords * wordWidth
                    # align end of frame to word
            endOfThisFrame = min(startOfThisFrame +
                                 maxFrameLen, endOfThisFrame)

            yield FrameTmpl(transaction,
                            wordWidth,
                            startOfThisFrame,
                            endOfThisFrame,
                            parts)
            parts = []
            startOfThisFrame = endOfThisFrame

        # final padding on the end
        while withPadding and startOfThisFrame < transaction.bitAddrEnd:
            endOfThisFrame = min(startOfThisFrame +
                                 maxFrameLen, transaction.bitAddrEnd)

            yield FrameTmpl(transaction,
                            wordWidth,
                            startOfThisFrame,
                            endOfThisFrame,
                            [])

            startOfThisFrame = endOfThisFrame

    @internal
    def _wordIndx(self, addr: int):
        """
        convert bit address to index of word where this address is
        """
        return floor(addr / self.wordWidth)

    def getWordCnt(self):
        """
        Get count of words in this frame
        """
        return ceil((self.endBitAddr - self.startBitAddr) / self.wordWidth)

    def walkWords(self, showPadding: bool=False):
        """
        Walk enumerated words in this frame

        :attention: not all indexes has to be present, only words
            with items will be generated when not showPadding
        :param showPadding: padding TransParts are also present
        :return: generator of tuples (wordIndex, list of TransParts
            in this word)
        """
        wIndex = 0
        lastEnd = self.startBitAddr
        parts = []
        for p in self.parts:
            end = p.startOfPart
            if showPadding and end != lastEnd:
                # insert padding before data
                while end != lastEnd:
                    assert end >= lastEnd, (end, lastEnd)
                    endOfWord = ceil(
                        (lastEnd + 1) / self.wordWidth) * self.wordWidth
                    endOfPadding = min(endOfWord, end)
                    _p = TransPart(self, None, False, lastEnd, endOfPadding, 0)
                    parts.append(_p)

                    if endOfPadding >= endOfWord:
                        yield (wIndex, parts)
                        wIndex += 1
                        parts = []

                    lastEnd = endOfPadding

            if self._wordIndx(lastEnd) != self._wordIndx(p.startOfPart):
                # if input data continues to a next word, yield current word
                # and start processing next one
                yield (wIndex, parts)

                wIndex += 1
                parts = []
                lastEnd = p.endOfPart

            parts.append(p)
            lastEnd = p.endOfPart
            if lastEnd % self.wordWidth == 0:
                # if we can not add anything to this word,
                # yield it directly and continue on next word
                yield (wIndex, parts)

                wIndex += 1
                parts = []

        if showPadding and (parts
                            or lastEnd != self.endBitAddr
                            or lastEnd % self.wordWidth != 0):
            # align end to end of last word
            end = ceil(self.endBitAddr / self.wordWidth) * self.wordWidth
            # padding is non removable if it is part of data
            # and it is removable if it was generated by frame alignment
            endOfNonRemovablePadding = self.origin.bitAddrEnd
            while end != lastEnd:
                assert end >= lastEnd, (end, lastEnd)
                endOfWord = ((lastEnd // self.wordWidth) + 1) * self.wordWidth
                endOfPadding = min(endOfWord, end)
                if lastEnd < endOfNonRemovablePadding:
                    endOfPadding = min(endOfPadding, endOfNonRemovablePadding)
                    can_be_removed = False
                else:
                    can_be_removed = True
                _p = TransPart(self, None, can_be_removed,
                               lastEnd, endOfPadding, 0)
                _p.parent = self
                parts.append(_p)

                if endOfPadding >= endOfWord:
                    yield (wIndex, parts)
                    wIndex += 1
                    parts = []

                lastEnd = endOfPadding

        if parts:
            # in the case end of frame is not aligned to end of word
            yield (wIndex, parts)

    @staticmethod
    def fieldToDataDict(dtype, data, res):
        return FrameTmpl._fieldToDataDict(dtype, (dtype,), data, res)

    @staticmethod
    def _fieldToDataDict(dtype, path, data, res):
        """
        Construct dictionary {StructField:value} for faster lookup of values
        for fields
        """
        # assert data is None or isinstance(data, dict)
        for f in dtype.fields:
            try:
                fVal = data[f.name]
            except KeyError:
                fVal = None

            new_path = (*path, f)
            if isinstance(f.dtype, Bits):
                if fVal is not None:
                    assert isinstance(fVal, int)
                    res[new_path] = fVal
            elif isinstance(f.dtype, HStruct):
                if fVal:
                    FrameTmpl._fieldToDataDict(f.dtype, new_path, fVal, res)
            elif isinstance(f.dtype, HArray):
                if fVal:
                    # assert isinstance(fVal, class_or_tuple)
                    res[new_path] = fVal

        return res

    def packData(self, data):
        """
        Pack data into list of BitsVal of specified dataWidth

        :param data: dict of values for struct fields {fieldName: value}

        :return: list of BitsVal which are representing values of words
        """
        typeOfWord = Bits(self.wordWidth, None)
        fieldToVal = self._fieldToTPart
        if fieldToVal is None:
            fieldToVal = self._fieldToTPart = self.fieldToDataDict(
                self.origin.dtype,
                data,
                {})

        for _, transParts in self.walkWords(showPadding=True):
            # build a single data word
            actualVldMask = 0
            actualVal = 0
            for tPart in transParts:
                high, low = tPart.getBusWordBitRange()
                fhigh, flow = tPart.getFieldBitRange()
                if not tPart.isPadding:
                    val = fieldToVal.get(tPart.tmpl.origin, None)
                else:
                    val = None

                if val is None:
                    newBits = 0
                    vld = 0
                else:
                    newBits = get_bit_range(val, flow, fhigh - flow)
                    vld = mask(high - low)

                actualVal = set_bit_range(actualVal, low, high - low, newBits)
                actualVldMask = set_bit_range(actualVldMask, low, high - low, vld)

            v = typeOfWord.getValueCls()(typeOfWord, actualVal,
                                         actualVldMask)
            yield v

    @internal
    def __repr__getName(self, transPart: TransPart, fieldWidth: int):
        """
        Get name string for a field
        """
        if transPart.isPadding:
            return "X" * fieldWidth
        else:
            path = transPart.tmpl.getFieldPath()
            names = []
            for p in path:
                if isinstance(p, int):
                    names.append(f"[{p:d}]")
                else:
                    if names:
                        names.append(f".{p:s}")
                    else:
                        names.append(p)

            return "".join(names)

    @internal
    def __repr__word(self,
                     index: int,
                     width: int,
                     padding: int,
                     transParts: List[TransPart]):
        buff = [f"{index: <{padding}}|", ]
        DW = self.wordWidth
        partsWithChoice = []

        endAlignment = transParts[-1].endOfPart % DW
        if endAlignment:
            # -1 for ending |
            percentOfWidth = (DW - endAlignment) / DW
            # -1 for ending |
            fieldWidth = max(0, int(percentOfWidth * width) - 1)
            assert fieldWidth >= 0
            s = '%s|' % ("^" * fieldWidth)
            buff.append(s)

        for tp in reversed(transParts):
            percentOfWidth = tp.bit_length() / DW
            # -1 for ending |
            fieldWidth = max(0, int(percentOfWidth * width) - 1)
            assert fieldWidth >= 0

            # percentOffset = (tp.inFrameBitAddr % DW) / DW
            # offset = int(percentOffset * width)
            if isinstance(tp, ChoicesOfFrameParts):
                name = "<union>"
                partsWithChoice.append(tp)
            else:
                name = self.__repr__getName(tp, fieldWidth)
            buff.append(f'{name: ^{fieldWidth}}|')
        return ("".join(buff), partsWithChoice)

    def __repr__(self, scale=1):
        buff = []
        s = f"<{self.__class__.__name__:s} start:{self.startBitAddr:d}, end:{self.endBitAddr:d}"
        buff.append(s)

        padding = 5
        DW = self.wordWidth
        width = int(DW * scale)

        buff.append(
            '{0: <{padding}}{1: <{halfLineWidth}}{2: >{halfLineWidth}}'.format(
                "", DW - 1, 0, padding=padding, halfLineWidth=width // 2))
        line = '{0: <{padding}}{1:-<{lineWidth}}'.format(
            "", "", padding=padding, lineWidth=width + 1)
        buff.append(line)

        for w, transParts in self.walkWords(showPadding=True):
            wStr, partsWithChoice = self.__repr__word(
                w, width, padding, transParts)
            buff.append(wStr)
            while partsWithChoice:
                for parts in zip_longest(*partsWithChoice):
                    parts = list(flatten(parts, level=1))
                    wStr, _partsWithChoice = self.__repr__word(
                        w, width, padding, parts)
                    buff.append(wStr)
                partsWithChoice = _partsWithChoice

        buff.append(line)
        buff.append(">")
        return "\n".join(buff)



from copy import copy
from typing import Optional, List

from hdlConvertorAst.hdlAst._expr import HdlValueId, HdlValueInt, HdlOp, \
    HdlOpType
from hdlConvertorAst.hdlAst._statements import HdlStmIf, HdlStmAssign, \
    HdlStmProcess, HdlStmBlock
from hdlConvertorAst.hdlAst._structural import HdlModuleDec, HdlModuleDef
from hdlConvertorAst.to.basic_hdl_sim_model.keywords import SIMMODEL_KEYWORDS
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model.utils import hdl_getattr, \
    hdl_map_asoc, hdl_call
from hdlConvertorAst.translate.common.name_scope import LanguageKeyword, NameScope
from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import AllOps
from hwt.hdl.portItem import HdlPortItem
from hwt.hdl.statements.assignmentContainer import HdlAssignmentContainer
from hwt.hdl.statements.codeBlockContainer import HdlStmCodeBlockContainer
from hwt.hdl.statements.ifContainter import IfContainer
from hwt.hdl.statements.switchContainer import SwitchContainer
from hwt.serializer.generic.constant_cache import ConstantCache
from hwt.serializer.generic.to_hdl_ast import ToHdlAst
from hwt.serializer.simModel.tmpVarConstructorConstOnly import TmpVarConstructorConstOnly
from hwt.serializer.simModel.types import ToHdlAstSimModel_types
from hwt.serializer.simModel.value import ToHdlAstSimModel_value
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwtSimApi.basic_hdl_simulator.sim_utils import sim_eval_cond


class ToHdlAstSimModel(ToHdlAstSimModel_value, ToHdlAstSimModel_types,
                       ToHdlAst):
    """
    Serializer which converts :class:`hwt.synthesizer.unit.Unit` instances to simulator code
    """
    _keywords_dict = {kw: LanguageKeyword() for kw in SIMMODEL_KEYWORDS}
    SIM_EVAL_COND = HdlValueId("sim_eval_cond", obj=sim_eval_cond)
    C = HdlValueId("c", obj=LanguageKeyword())
    CVLD = HdlValueId("cVld", obj=LanguageKeyword())
    TMP_VAR_CONSTRUCTOR = TmpVarConstructorConstOnly

    def __init__(self, name_scope: Optional[NameScope]=None):
        super(ToHdlAstSimModel, self).__init__(name_scope)
        self.currentUnit = None
        self.stm_outputs = {}

    def as_hdl_HdlModuleDec(self, o: HdlModuleDec):
        # convert types, exprs
        # delete params because they should not be used in expressions and thus
        # are useless
        new_o = copy(o)
        new_o.params = []
        new_o.ports = [self.as_hdl_HdlPortItem(p) for p in o.ports]
        return new_o

    def as_hdl_PortConnection(self, o: HdlPortItem):
        assert isinstance(o, HdlPortItem), o
        intern, outer = o.getInternSig(), o.getOuterSig()
        assert not intern.hidden, intern
        assert not outer.hidden, outer
        intern_hdl = HdlValueId(intern.name, obj=intern)
        outer_hdl = HdlValueId(outer.name, obj=outer)
        pm = hdl_map_asoc(intern_hdl, outer_hdl)
        return pm

    def as_hdl_HdlAssignmentContainer(self, a: HdlAssignmentContainer):
        dst, dst_indexes, src = self._as_hdl_HdlAssignmentContainer_auto_conversions(a)
        ev = HdlValueInt(int(a._event_dependent_from_branch == 0), None, None)
        if dst_indexes is not None:
            src = (src, dst_indexes, ev)
        else:
            src = (src, ev)
        hdl_dst = hdl_getattr(hdl_getattr(self.SELF_IO, dst.name), "val_next")
        hdl_a = HdlStmAssign(src, hdl_dst)
        hdl_a.is_blocking = dst.virtual_only
        return hdl_a

    def as_hdl_IfContainer_out_invalidate_section(self,
                                                  outputs: List[RtlSignalBase],
                                                  parent: IfContainer):
        outputInvalidateStms = []
        for o in outputs:
            # [TODO] look up indexes
            indexes = None
            v = o._dtype.from_py(None)
            oa = HdlAssignmentContainer(v, o, indexes,
                            virtual_only=True, parentStm=parent,
                            event_dependent_from_branch=parent._event_dependent_from_branch)
            outputInvalidateStms.append(self.as_hdl_HdlAssignmentContainer(oa))

        if len(outputInvalidateStms) == 1:
            return outputInvalidateStms[0]
        else:
            b = HdlStmBlock()
            b.body = outputInvalidateStms
            return b

    def as_hdl_IfContainer_cond_eval(self, cond):
        """
        constructs condition evaluation statement
        c, cVld = sim_eval_cond(cond)
        """
        c, cVld = self.C, self.CVLD
        cond = self.as_hdl_cond(cond, True)
        cond_eval = hdl_call(self.SIM_EVAL_COND, [cond])
        cond_eval = HdlStmAssign(cond_eval, (c, cVld))
        cond_eval.is_blocking = True
        return c, cVld, cond_eval

    def as_hdl_IfContainer(self, ifc: IfContainer) -> HdlStmIf:
        """
        .. code-block:: python

            if cond:
                ...
            else:
                ...

        will become

        .. code-block:: python

            c, cVld = sim_eval_cond(cond)
            if not cVld:
                # ivalidate outputs
            elif c:
                ... # original if true branch
            else:
                ... # original if else brach
        """
        invalidate_block = self.as_hdl_IfContainer_out_invalidate_section(
            ifc._outputs, ifc)
        c, cVld, cond_eval = self.as_hdl_IfContainer_cond_eval(ifc.cond)
        _if = HdlStmIf()
        res = HdlStmBlock()
        res.body = [cond_eval, _if]

        _if.cond = HdlOp(HdlOpType.NEG_LOG, [cVld, ])
        _if.if_true = invalidate_block

        if_true = self.as_hdl_statements(ifc.ifTrue)
        _if.elifs.append((c, if_true))
        elifs = iter(ifc.elIfs)
        for eif_c, eif_stms in elifs:
            c, cVld, cond_eval = self.as_hdl_IfContainer_cond_eval(eif_c)
            newIf = HdlStmIf()
            newIf.cond = HdlOp(HdlOpType.NEG_LOG, [cVld, ])
            newIf.if_true = invalidate_block

            if_true = self.as_hdl_statements(eif_stms)
            newIf.elifs.append((c, if_true))

            _if.if_false = HdlStmBlock()
            _if.if_false.body = [cond_eval, newIf]
            _if = newIf

        _if.if_false = self.as_hdl_statements(ifc.ifFalse)

        return res

    def as_hdl_SwitchContainer(self, sw: SwitchContainer) -> HdlStmIf:
        "switch -> if"
        switchOn = sw.switchOn

        def mkCond(c):
            return switchOn._eq(c)

        elIfs = []

        for key, statements in sw.cases[1:]:
            elIfs.append((mkCond(key), statements))
        ifFalse = sw.default

        topCond = mkCond(sw.cases[0][0])
        topIf = IfContainer(topCond,
                            ifTrue=sw.cases[0][1],
                            ifFalse=ifFalse,
                            elIfs=elIfs)

        topIf._sensitivity = sw._sensitivity
        topIf._inputs = sw._inputs
        topIf._outputs = sw._outputs

        return self.as_hdl_IfContainer(topIf)

    def sensitivityListItem(self, item, anyEventDependent):
        if isinstance(item, Operator):
            op = item.operator
            if op == AllOps.RISING_EDGE:
                sens = HdlOpType.RISING
            elif op == AllOps.FALLING_EDGE:
                sens = HdlOpType.FALLING
            else:
                raise TypeError("This is not an event sensitivity", op)

            return HdlOp(sens, [HdlValueId(item.operands[0].name)])
        else:
            return HdlValueId(item.name)

    def has_to_be_process(self, proc):
        return True

    def can_pop_process_wrap(self, statements, hasToBeVhdlProcess):
        return False

    def as_hdl_HdlStmCodeBlockContainer(self, proc: HdlStmCodeBlockContainer) -> HdlStmProcess:
        p = ToHdlAst.as_hdl_HdlStmCodeBlockContainer(self, proc)
        self.stm_outputs[p] = sorted(
            [HdlValueId(i.name, obj=i)
             for i in proc._outputs]
        )
        return p

    def as_hdl_extraVarsInit(self, extraVars):
        return []

    def _as_hdl_HdlModuleDef_body(self, *args) -> HdlModuleDef:
        orig_const_cache = self.constCache
        try:
            self.constCache = ConstantCache(self, self.tmpVars)
            return ToHdlAst._as_hdl_HdlModuleDef_body(self, *args)
        finally:
            self.constCache = orig_const_cache



from math import inf, isinf
from typing import List, Optional

from hwt.doc_markers import internal
from hwt.hdl.types.hdlType import HdlType
from hwt.serializer.generic.indent import getIndent


class HStream(HdlType):
    """
    Stream is an abstract type. It is an array with unspecified size.

    :ivar ~.element_t: type of smalest chunk of data
        which can be send over this stream
    :ivar ~.len_min: minimum repetitions of element_t (inclusive interval)
    :ivar ~.len_max: maximum repetitions of element_t (inclusive interval)
    :ivar ~.start_offsets: list of numbers which represents the number of invalid bytes
        before valid data on stream (invalid bytes means the bytes
        which does not have bit validity set, e.g. Axi4Stream keep=0b10 -> offset=1
        )
    """

    def __init__(self, element_t,
                 frame_len=inf,
                 start_offsets: Optional[List[int]]=None,
                 const=False):
        super(HStream, self).__init__(const=const)
        self.element_t = element_t
        if isinstance(frame_len, float) and isinf(frame_len):
            frame_len = (1, inf)
        elif isinstance(frame_len, int):
            frame_len = (frame_len, frame_len)
        self.len_min, self.len_max = frame_len
        if start_offsets is None:
            start_offsets = (0, )
        self.start_offsets = tuple(start_offsets)

    def bit_length(self):
        if self.len_min != self.len_max or isinf(self.len_max):
            raise TypeError("This HStream does not have constant size", self)
        else:
            # len_min == len_max
            return self.len_min * self.element_t.bit_length()

    def __eq__(self, other: HdlType):
        if self is other:
            return True
        if (type(self) is type(other)):
            if self.start_offsets == other.start_offsets \
                    and self.len_min == other.len_min \
                    and self.len_max == other.len_max:
                return self.element_t == other.element_t
        return False

    def __hash__(self):
        return hash((self.start_offsets, self.len_min, self.len_max, self.element_t))

    @internal
    @classmethod
    def getValueCls(cls):
        try:
            return cls._valCls
        except AttributeError:
            from hwt.hdl.types.streamVal import HStreamVal
            cls._valCls = HStreamVal
            return cls._valCls

    def __repr__(self, indent=0, withAddr=None, expandStructs=False):
        return "%s<%s len:%s, align:%r\n%s>" % (
            getIndent(indent),
            self.__class__.__name__,
            (self.len_min, self.len_max),
            self.start_offsets,
            self.element_t.__repr__(indent=indent+1,
                                    withAddr=withAddr,
                                    expandStructs=expandStructs),
        )



from hwt.doc_markers import internal
from hwt.hdl.types.hdlType import HdlType


class HString(HdlType):

    def all_mask(self):
        return 1

    @internal
    @classmethod
    def getValueCls(cls):
        try:
            return cls._valCls
        except AttributeError:
            from hwt.hdl.types.stringVal import HStringVal
            cls._valCls = HStringVal
            return cls._valCls



from hwt.serializer.generic.tmpVarConstructor import TmpVarConstructor
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


class ConstantCache(object):
    """
    Container of constants for serializer.
    Used to extract constants as constant variables.
    """

    def __init__(self, toHdlAst, tmpVars: TmpVarConstructor):
        self.tmpVars = tmpVars
        self.toHdlAst = toHdlAst

        # {value:usedName}
        self._cache = {}

    def extract_const_val_as_const_var(self, val) -> RtlSignalBase:
        """
        Create a constant variable with a value specified
        or use existitng variable with same value
        """
        try:
            return self._cache[val]
        except KeyError:
            if isinstance(val.val, int):
                if val.val < 0:
                    name = "const_m%d_" % -val.val
                else:
                    name = f"const_{val.val:d}_"
            else:
                name = "const_"

            toHdlAst = self.toHdlAst
            cc = toHdlAst.constCache
            try:
                # dissable const cache as the value is beeing extracted
                # and we want to prevent recursion
                toHdlAst.constCache = None
                c = self.tmpVars.create_var(name, val._dtype, def_val=val, const=True)
            finally:
                toHdlAst.constCache = cc

            self._cache[val] = c
            return c



from hwt.doc_markers import internal
from hwt.hdl.constants import DIRECTION
from hwt.hdl.portItem import HdlPortItem
from hwt.hdl.variables import SignalItem


@internal
def portItemfromSignal(s: SignalItem, component, d: DIRECTION):
    return HdlPortItem(s.name, d, s._dtype, component)



from copy import copy
from itertools import chain
from typing import Tuple

from hdlConvertorAst.hdlAst._bases import iHdlStatement
from hdlConvertorAst.hdlAst._statements import HdlStmBlock
from hdlConvertorAst.hdlAst._structural import HdlModuleDef, HdlCompInst
from hwt.hdl.sensitivityCtx import SensitivityCtx
from hwt.hdl.statements.assignmentContainer import HdlAssignmentContainer
from hwt.hdl.statements.codeBlockContainer import HdlStmCodeBlockContainer
from hwt.hdl.statements.ifContainter import IfContainer
from hwt.hdl.statements.switchContainer import SwitchContainer
from hwt.serializer.combLoopAnalyzer.tarjan import StronglyConnectedComponentSearchTarjan
from hwt.serializer.resourceAnalyzer.analyzer import ResourceAnalyzer
from hwt.synthesizer.componentPath import ComponentPath
from hwt.synthesizer.unit import Unit
from ipCorePackager.constants import DIRECTION


def collect_comb_inputs(ctx, seen, input_signal, comb_inputs):
    if input_signal not in seen:
        seen.add(input_signal)
        input_signal._walk_sensitivity(comb_inputs, seen, ctx)


def collect_comb_drivers(path_prefix: Tuple[Unit, ...],
                         stm: iHdlStatement,
                         comb_connection_matrix: dict,
                         comb_inputs: tuple):
    if isinstance(stm, HdlAssignmentContainer):
        ctx = SensitivityCtx()
        seen = set()
        # merge condition inputs to current_comb_inputs
        current_comb_inputs = set(comb_inputs)
        for inp in stm._inputs:
            collect_comb_inputs(ctx, seen, inp, current_comb_inputs)

        for o in stm._outputs:
            o_key = path_prefix / o
            for i in current_comb_inputs:
                con = comb_connection_matrix.setdefault(path_prefix / i, set())
                con.add(o_key)

    elif isinstance(stm, IfContainer):
        current_comb_inputs = set(comb_inputs)  # intended copy
        elifs = ((stm.cond, stm.ifTrue), *stm.elIfs)
        ev_dep_branch = stm._event_dependent_from_branch
        ctx = SensitivityCtx()
        seen = set()
        found_event_dep = False
        for branch_i, (cond, stms) in enumerate(elifs):
            # [TODO] check if this works for clock gating
            if ev_dep_branch is not None and ev_dep_branch == branch_i:
                found_event_dep = True
                break

            collect_comb_inputs(ctx, seen, cond, current_comb_inputs)

            for sub_stm in stms:
                collect_comb_drivers(path_prefix, sub_stm, comb_connection_matrix, current_comb_inputs)

        if not found_event_dep and stm.ifFalse is not None:
            for sub_stm in stms:
                collect_comb_drivers(path_prefix, sub_stm, comb_connection_matrix, current_comb_inputs)

    elif isinstance(stm, HdlStmCodeBlockContainer):
        for sub_stm in stm.statements:
            collect_comb_drivers(path_prefix, sub_stm, comb_connection_matrix, comb_inputs)

    elif isinstance(stm, SwitchContainer):
        current_comb_inputs = set(comb_inputs)
        ctx = SensitivityCtx()
        seen = set()
        collect_comb_inputs(ctx, seen, stm.switchOn, current_comb_inputs)
        cases = stm.cases
        if stm.default is not None:
            cases = chain(cases, ((None, stm.default),))

        for (_, case_stms) in cases:
            for sub_stm in case_stms:
                collect_comb_drivers(path_prefix, sub_stm, comb_connection_matrix, current_comb_inputs)

    else:
        raise NotImplementedError(stm)


class CombLoopAnalyzer():
    """
    Visitor which can walk synthetized hwt :class:`hwt.synthesizer.unit.Unit` instances and detect clusters connected by combinational logic
    """

    def __init__(self):
        # RtlSignal: Set[RtlSignal]
        self.comb_connection_matrix = {}
        # RtlSignal: set of signals in comb loop
        self._report = {}
        self.actual_path_prefix = ComponentPath()

    def visit_Unit(self, u: Unit):
        if u._shared_component_with is None:
            arch = u._ctx.arch
        else:
            _u, _, _ = u._shared_component_with
            arch = _u._ctx.arch
        assert arch is not None, u

        self.visit_HdlModuleDef(arch)

    def visit_HdlModuleDef(self, m: HdlModuleDef) -> None:
        ResourceAnalyzer.visit_HdlModuleDef(self, m)

    def visit_HdlStmCodeBlockContainer(self, proc: HdlStmCodeBlockContainer) -> None:
        collect_comb_drivers(self.actual_path_prefix, proc, self.comb_connection_matrix, tuple())

    def visit_HdlCompInst(self, o: HdlCompInst) -> None:
        orig_path_prefix = self.actual_path_prefix
        in_component_path_prefix = orig_path_prefix
        if o.origin._shared_component_with is not None:
            in_component_path_prefix = in_component_path_prefix / o.origin


        try:
            assert o.origin, o
            self.actual_path_prefix = in_component_path_prefix
            self.visit_Unit(o.origin)

            for pm in o.port_map:
                if pm.direction == DIRECTION.OUT:
                    k = in_component_path_prefix / pm.src
                    v = orig_path_prefix / pm.dst
                elif pm.direction == DIRECTION.IN:
                    k = orig_path_prefix / pm.src
                    v = in_component_path_prefix / pm.dst
                else:
                    raise NotImplementedError(pm.direction)
                self.comb_connection_matrix.setdefault(k, set()).add(v)

        finally:
            self.actual_path_prefix = orig_path_prefix

    def report(self):
        scc_search = StronglyConnectedComponentSearchTarjan(self.comb_connection_matrix)
        for scc in scc_search.search_strongly_connected_components():
            if len(scc) > 1:
                yield scc



from typing import Union, List

from hwt.hdl.types.array import HArray
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.hdlType import HdlType
from hwt.hdl.types.stream import HStream
from hwt.hdl.types.struct import HStruct
from hwt.hdl.types.typeCast import toHVal
from hwt.hdl.types.union import HUnion
from hwt.hdl.value import HValue
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


def walkFlattenFields(sigOrVal: Union[RtlSignalBase, HValue], skipPadding=True):
    """
    Walk all simple values in HStruct or HArray
    """
    t = sigOrVal._dtype
    if isinstance(t, Bits):
        yield sigOrVal
    elif isinstance(t, HUnion):
        yield from walkFlattenFields(sigOrVal._val, skipPadding=skipPadding)
    elif isinstance(t, HStruct):
        for f in t.fields:
            isPadding = f.name is None
            if not isPadding or not skipPadding:
                if isPadding:
                    v = f.dtype.from_py(None)
                else:
                    v = getattr(sigOrVal, f.name)

                yield from walkFlattenFields(v)

    elif isinstance(t, HArray):
        for item in sigOrVal:
            yield from walkFlattenFields(item)
    elif isinstance(t, HStream):
        assert isinstance(sigOrVal, HValue), sigOrVal
        for v in sigOrVal:
            yield from walkFlattenFields(v)
    else:
        raise NotImplementedError(t)


def HdlValue_unpack(t: HdlType,
                    data: List[Union[HValue, RtlSignalBase, int]],
                    getDataFn=None, dataWidth=None):
    """
    Parse raw Bits array to a value of specified HdlType
    """
    if getDataFn is None:
        assert dataWidth is not None

        def _getDataFn(x):
            return toHVal(x)._auto_cast(Bits(dataWidth))

        getDataFn = _getDataFn

    val = t.from_py(None)

    fData = iter(data)

    # actual is storage variable for items from frameData
    actualOffset = 0
    actual = None

    for v in walkFlattenFields(val, skipPadding=False):
        # walk flatten fields and take values from fData and parse them to
        # field
        required = v._dtype.bit_length()

        if actual is None:
            actualOffset = 0
            try:
                actual = getDataFn(next(fData))
            except StopIteration:
                raise ValueError("Input data too short")

            if dataWidth is None:
                dataWidth = actual._dtype.bit_length()
            actuallyHave = dataWidth
        else:
            actuallyHave = actual._dtype.bit_length() - actualOffset

        while actuallyHave < required:
            # collect data for this field
            try:
                d = getDataFn(next(fData))
            except StopIteration:
                raise ValueError("Input data too short")

            actual = d._concat(actual)
            actuallyHave += dataWidth

        if actuallyHave >= required:
            # parse value of actual to field
            # skip padding
            _v = actual[(required + actualOffset):actualOffset]
            _v = _v._auto_cast(v._dtype)
            v.val = _v.val
            v.vld_mask = _v.vld_mask

            # update slice out what was taken
            actuallyHave -= required
            actualOffset += required

        if actuallyHave == 0:
            actual = None

    if actual is not None:
        assert actual._dtype.bit_length(
        ) - actualOffset < dataWidth, (
            "It should be just a padding at the end of frame"
        )
    return val


def is_only_padding(t: HdlType):
    if isinstance(t, HStruct):
        for f in t.fields:
            if f.name is not None and not is_only_padding(f.dtype):
                return False
        return True
    elif isinstance(t, (HArray, HStream)):
        return is_only_padding(t.element_t)
    return False



from hdlConvertorAst.hdlAst._expr import HdlOp, HdlOpType, HdlValueId
from hdlConvertorAst.to.verilog.constants import SIGNAL_TYPE
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model.utils import hdl_index
from hdlConvertorAst.translate.common.name_scope import LanguageKeyword
from hwt.hdl.types.array import HArray
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.enum import HEnum
from hwt.hdl.types.hdlType import MethodNotOverloaded
from hwt.serializer.verilog.types import ToHdlAstVerilog_types
from hwt.hdl.types.defs import FLOAT64
from hwt.hdl.types.float import HFloat


class ToHdlAstSystemC_type():
    sc_int = HdlValueId("sc_int", obj=LanguageKeyword())
    sc_uint = HdlValueId("sc_uint", obj=LanguageKeyword())
    sc_bigint = HdlValueId("sc_bigint", obj=LanguageKeyword())
    sc_biguint = HdlValueId("sc_biguint", obj=LanguageKeyword())
    sc_signal = HdlValueId("sc_signal", obj=LanguageKeyword())
    STRING = HdlOp(HdlOpType.DOUBLE_COLON, HdlValueId("string", obj=LanguageKeyword()))

    def does_type_requires_extra_def(self, t, other_types):
        try:
            return t._as_hdl_requires_def(self, other_types)
        except MethodNotOverloaded:
            pass
        return False

    def as_hdl_HdlType_str(self, typ, declaration=False):
        assert not declaration
        return self.STRING

    def as_hdl_HdlType_bits(self, typ: Bits, declaration=False):
        if declaration:
            raise NotImplementedError()

        w = typ.bit_length()

        if w <= 64:
            if typ.signed:
                typeBaseName = self.sc_int
            else:
                typeBaseName = self.sc_uint
        else:
            if typ.signed:
                typeBaseName = self.sc_bigint
            else:
                typeBaseName = self.sc_biguint

        t = HdlOp(HdlOpType.PARAMETRIZATION,
                    [typeBaseName, self.as_hdl_int(w)])
        if self.signalType == SIGNAL_TYPE.WIRE:
            t = HdlOp(HdlOpType.PARAMETRIZATION, [self.sc_signal, t])
        return t

    def as_hdl_HdlType_array(self, typ: HArray, declaration=False):
        if declaration:
            return super(ToHdlAstSystemC_type, self).as_hdl_HdlType_array(
                self, typ, declaration=declaration)
        else:
            _int = self.as_hdl_int
            size = _int(int(typ.size))
            return hdl_index(self.as_hdl_HdlType(typ.element_t), size)

    def as_hdl_HdlType_enum(self, typ: HEnum, declaration=False):
        return ToHdlAstVerilog_types.as_hdl_HdlType_enum(
            self, typ, declaration=declaration)

    def as_hdl_HdlType_float(self, typ: HFloat, declaration=False):
        if typ == FLOAT64:
            return HdlValueId("double")
        else:
            raise NotImplementedError(typ)



from hwt.serializer.generic.tmpVarConstructor import TmpVarConstructor
from hwt.synthesizer.rtlLevel.rtlSignal import RtlSignal


class TmpVarConstructorConstOnly(TmpVarConstructor):

    def finish_var_init(self, var: RtlSignal):
        hdl = self.extraVarsHdl
        as_hdl = self.toHdlAst.as_hdl_SignalItem(var, declaration=True)
        hdl.append(as_hdl)

    def sort_hdl_declarations_first(self):
        pass  # always sorted



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-

from os import path
from setuptools import setup, find_packages

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(name="hwt",
      version="3.8",
      description="hdl synthesis toolkit",
      long_description=long_description,
      long_description_content_type="text/markdown",
      url="https://github.com/Nic30/hwt",
      author="Michal Orsak",
      author_email="Nic30original@gmail.com",
      classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)",
        "Topic :: System :: Hardware",
        "Topic :: System :: Emulators",
        "Topic :: Utilities"],
      install_requires=[
          "hdlConvertorAst>=1.0", # conversions to SystemVerilog, VHDL
          "ipCorePackager>=0.6",  # generator of IPcore packages (IP-xact, ...)
          "hwtSimApi>=1.3",  # simulator API
          "pyDigitalWaveTools>=1.1",  # simulator output dump
      ],
      license="MIT",
      packages=find_packages(),
      zip_safe=True
)



"""
Serializer mode specifies if hdl objects derived from parent unit
should be serialized to target HDL or not

use serialize* methods to specify serialization mode for unit class

.. code-block:: python

    @serializeExclude
    class MyUnit(Unit):
        # ...
        pass

"""

from collections import namedtuple

from hwt.doc_markers import internal


@internal
def freeze_dict(data):
    keys = sorted(data.keys())
    if keys:
        frozen_type = namedtuple(''.join(keys), keys)
    else:
        return tuple()
    return frozen_type(**data)


@internal
def paramsToValTuple(unit):
    # [TODO] check sub params
    d = {}
    for p in unit._params:
        v = p.get_value()
        d[p._name] = v
    return freeze_dict(d)


def serializeExclude(cls):
    """
    Never serialize HDL objects from this class
    """
    cls._serializeDecision = staticmethod(_serializeExclude_eval)
    return cls


def serializeOnce(cls):
    """
    Serialize HDL objects only once per class
    """
    cls._serializeDecision = staticmethod(_serializeOnce_eval)
    return cls


def serializeParamsUniq(cls):
    """
    Decide to serialize only when parameters are unique
    """
    cls._serializeDecision = staticmethod(_serializeParamsUniq_eval)
    return cls


@internal
def _serializeExclude_eval(parentUnit, priv):
    """
    Always decide not to serialize obj

    :param priv: private data for this function first unit of this class
    :return: tuple (do serialize this object, next priv, replacement unit)
    """

    # do not use this :class:`hwt.synthesizer.unit.Unit` instance and do not use any prelacement
    # (usefull when the :class:`hwt.synthesizer.unit.Unit` instance is a placeholder for something
    #  which already exists in hdl word)
    if priv is None:
        priv = parentUnit
        return False, priv, None
    else:
        return False, priv, priv


@internal
def _serializeOnce_eval(parentUnit, priv):
    """
    Decide to serialize only first obj of it's class

    :param priv: private data for this function
        (first object with class == obj.__class__)

    :return: tuple (do serialize this object, next priv, replacement unit)
        where priv is private data for this function
        (first object with class == obj.__class__)
    """
    if priv is None:
        priv = parentUnit
        serialize = True
        replacement = None
        # use this :class:`hwt.synthesizer.unit.Unit` instance and store it for later use
    else:
        # use existing :class:`hwt.synthesizer.unit.Unit` instance
        serialize = False
        replacement = priv

    return serialize, priv, replacement


@internal
def _serializeParamsUniq_eval(parentUnit, priv):
    """
    Decide to serialize only objs with uniq parameters and class

    :param priv: private data for this function
        ({frozen_params: obj})

    :return: tuple (do serialize this object, next priv, replacement unit)
    """

    params = paramsToValTuple(parentUnit)

    if priv is None:
        priv = {}

    try:
        prevUnit = priv[params]
    except KeyError:
        priv[params] = parentUnit
        # serialize new
        return True, priv, None

    # use previous :class:`hwt.synthesizer.unit.Unit` instance with same config
    return False, priv, prevUnit



from hdlConvertorAst.hdlAst._expr import HdlOp, HdlValueId
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model.utils import hdl_getattr, \
    hdl_call
from hdlConvertorAst.translate.common.name_scope import LanguageKeyword
from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import AllOps
from hwt.serializer.hwt.context import ValueWidthRequirementScope
from hwt.serializer.simModel.value import ToHdlAstSimModel_value


class ToHdlAstHwt_ops():
    CONCAT = HdlValueId("Concat", obj=LanguageKeyword())
    op_transl_dict = ToHdlAstSimModel_value.op_transl_dict
    _cast_ops = ToHdlAstSimModel_value._cast_ops

    def as_hdl_Operator(self, op: Operator):
        ops = op.operands
        o = op.operator

        with ValueWidthRequirementScope(self, o == AllOps.CONCAT):
            if o in self._cast_ops:
                op0 = hdl_getattr(self.as_hdl(ops[0]), "_reinterpret_cast")
                op1 = self.as_hdl_HdlType(op.result._dtype)
                return hdl_call(op0, [op1, ])
            elif o == AllOps.EQ:
                return hdl_call(hdl_getattr(self.as_hdl(ops[0]), "_eq"),
                                [self.as_hdl(ops[1])])
            elif o == AllOps.CONCAT:
                return hdl_call(self.CONCAT,
                                [self.as_hdl(o2) for o2 in ops])
            elif o == AllOps.TERNARY:
                cond, op0, op1 = ops
                cond = self.as_hdl(cond)
                with ValueWidthRequirementScope(self, True):
                    op0 = self.as_hdl(op0)
                    op1 = self.as_hdl(op1)
                return hdl_call(hdl_getattr(cond, "_ternary"), [op0, op1])
            else:
                o = self.op_transl_dict[o]
                return HdlOp(o, [self.as_hdl(o2)
                                 for o2 in ops])



"""
Commonly used constants during HW devel.
"""
from enum import Enum

# import constants from other packages to have them on one place
from ipCorePackager.constants import INTF_DIRECTION, DIRECTION
from hwtSimApi.constants import Time, CLK_PERIOD

READ = "READ"
WRITE = "WRITE"
READ_WRITE = "RW"
NOP = "NOP"


class SENSITIVITY(Enum):
    """
    Sensitivity used in sensitivity resolver
    """
    ANY = 0b11
    RISING = 0b01
    FALLING = 0b10



from hdlConvertorAst.hdlAst import HdlValueId, HdlValueInt
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model.utils import hdl_call, \
    hdl_getattr
from hdlConvertorAst.translate.common.name_scope import LanguageKeyword
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.slice import HSlice
from pyMathBitPrecise.bits3t import Bits3t


class ToHdlAstSimModel_types():
    """
    part of ToHdlAstSimModel responsible for type serialization
    """
    SELF = HdlValueId("self", obj=LanguageKeyword())
    BITS3T = HdlValueId("Bits3t", obj=Bits3t)
    SLICE = HdlValueId("slice", obj=slice)

    def as_hdl_HdlType_bits(self, typ: Bits, declaration=False):
        assert not declaration
        w = typ.bit_length()
        if isinstance(w, int):
            pass
        else:
            w = int(w)

        return hdl_call(self.BITS3T, [HdlValueInt(w, None, None),
                                      HdlValueInt(int(bool(typ.signed)), None, None)])

    def as_hdl_HdlType_slice(self, typ: HSlice, declaration=False):
        if declaration:
            raise NotImplementedError()
        else:
            return self.SLICE

    def as_hdl_HdlType_array(self, typ, declaration=False):
        if declaration:
            return super(ToHdlAstSimModel_types, self).as_hdl_HdlType_array(typ, declaration=declaration)
        else:
            t_name = self.name_scope.get_object_name(typ)
            return hdl_getattr(self.SELF, t_name)

    def as_hdl_HdlType_enum(self, typ, declaration=False):
        if declaration:
            return super(ToHdlAstSimModel_types, self).as_hdl_HdlType_enum(typ, declaration=True)
        else:
            t_name = self.name_scope.get_object_name(typ)
            return hdl_getattr(self.SELF, t_name)



from hdlConvertorAst.hdlAst._expr import HdlValueInt, HdlOpType, \
    HdlOp
from hdlConvertorAst.to.hdlUtils import bit_string
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model.utils import hdl_downto, \
    hdl_call
from hwt.doc_markers import internal
from hwt.hdl.operator import Operator
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.bitsVal import BitsVal
from hwt.hdl.types.defs import BOOL, BIT
from hwt.hdl.types.enumVal import HEnumVal
from hwt.hdl.types.sliceVal import HSliceVal
from hwt.hdl.value import HValue
from hwt.serializer.generic.ops import HWT_TO_HDLCONVEROTR_OPS
from hwt.serializer.generic.value import ToHdlAst_Value
from hwt.serializer.verilog.context import SignalTypeSwap
from hwt.serializer.verilog.utils import verilogTypeOfSig
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


class ToHdlAstVerilog_Value(ToHdlAst_Value):

    # TRUE = HdlValueId("true", obj=LanguageKeyword())
    # FALSE = HdlValueId("false", obj=LanguageKeyword())

    def as_hdl_BoolVal(self, val: BitsVal):
        return self.as_hdl_int(val.val)

    def as_hdl_cond(self, c, forceBool):
        assert isinstance(c, (RtlSignalBase, HValue))
        if not forceBool or c._dtype == BOOL:
            return self.as_hdl(c)
        elif c._dtype == BIT:
            return self.as_hdl(c)
        elif isinstance(c._dtype, Bits):
            return self.as_hdl(c != 0)
        else:
            raise NotImplementedError()

    def as_hdl_HEnumVal(self, val: HEnumVal):
        i = val._dtype._allValues.index(val.val)
        assert i >= 0
        return HdlValueInt(i, None, None)

    def as_hdl_SignalItem(self, si, declaration=False):
        if declaration:
            with SignalTypeSwap(self, verilogTypeOfSig(si)):
                return ToHdlAst_Value.as_hdl_SignalItem(self, si, declaration=declaration)
        else:
            return ToHdlAst_Value.as_hdl_SignalItem(self, si, declaration=declaration)

    def as_hdl_HSliceVal(self, val: HSliceVal):
        upper = val.val.start - 1
        return hdl_downto(self.as_hdl_Value(upper),
                          self.as_hdl_Value(val.val.stop))

    def sensitivityListItem(self, item, anyIsEventDependent):
        if isinstance(item, Operator):
            return HdlOp(HWT_TO_HDLCONVEROTR_OPS[item.operator],
                           [self.as_hdl(item.operands[0]), ])
        elif anyIsEventDependent:
            if item._dtype.negated:
                op = HdlOpType.FALLING
            else:
                op = HdlOpType.RISING
            return HdlOp(op, [self.as_hdl(item), ])

        return self.as_hdl(item)

    def as_hdl_HArrayVal(self, val):
        raise ValueError(
            "Verilog do not have a array constants(they are part of SV)"
            " and that is why array constants should converted to initialization"
            " in initial processes")

    @internal
    def as_hdl_BitString(self, v, width, force_vector, vld_mask, signed):
        v = bit_string(v, width, vld_mask=vld_mask)
        if signed:
            return hdl_call(self.SIGNED, [v, ])
        else:
            return v



from typing import List

from hwt.hdl.statements.statement import HdlStatement


def isSameStatementList(stmListA: List[HdlStatement],
                        stmListB: List[HdlStatement]) -> bool:
    """
    :return: True if two lists of HdlStatement instances are same
    """
    if stmListA is stmListB:
        return True
    if stmListA is None or stmListB is None:
        return False

    for a, b in zip(stmListA, stmListB):
        if not a.isSame(b):
            return False

    return True


def statementsAreSame(statements: List[HdlStatement]) -> bool:
    """
    :return: True if all statements are same
    """
    iterator = iter(statements)
    try:
        first = next(iterator)
    except StopIteration:
        return True

    return all(first.isSame(rest) for rest in iterator)



from datetime import datetime
import importlib
from io import StringIO
import os
import sys
from types import ModuleType
from typing import Union, Optional, Set, Tuple, Callable

from hwt.doc_markers import internal
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.enum import HEnum
from hwt.hdl.value import HValue
from hwt.serializer.serializer_filter import SerializerFilterDoNotExclude
from hwt.serializer.simModel import SimModelSerializer
from hwt.serializer.store_manager import SaveToStream, SaveToFilesFlat
from hwt.synthesizer.dummyPlatform import DummyPlatform
from hwt.synthesizer.interface import Interface
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.synthesizer.unit import Unit
from hwt.synthesizer.utils import to_rtl
from pyDigitalWaveTools.vcd.common import VCD_SIG_TYPE
from pyDigitalWaveTools.vcd.value_format import VcdBitsFormatter, \
    VcdEnumFormatter
from pyDigitalWaveTools.vcd.writer import VcdVarWritingScope, \
    VarAlreadyRegistered
from pyMathBitPrecise.bits3t import Bits3t
from pyMathBitPrecise.enum3t import Enum3t
from hwtSimApi.basic_hdl_simulator.model import BasicRtlSimModel
from hwtSimApi.basic_hdl_simulator.proxy import BasicRtlSimProxy
from hwtSimApi.basic_hdl_simulator.rtlSimulator import BasicRtlSimulator
from hwtSimApi.basic_hdl_simulator.sim_utils import ValueUpdater, \
    ArrayValueUpdater


class BasicRtlSimulatorWithSignalRegisterMethods(BasicRtlSimulator):
    supported_type_classes = tuple()

    def __init__(self, model_cls, synthesised_unit):
        """
        Only store variables for later construction
        """
        self.model_cls = model_cls
        self.synthesised_unit = synthesised_unit
        self.wave_writer = None
        self._obj2scope = {}
        self._traced_signals = set()

    def __call__(self) -> "BasicRtlSimulatorVcd":
        """
        Create and initialize the BasicRtlSimulatorWithVCD object
        """
        sim = self.__class__(self.model_cls, self.synthesised_unit)
        super(BasicRtlSimulatorWithSignalRegisterMethods, sim).__init__()
        model = self.model_cls(sim)
        model._init_body()
        sim.bound_model(model)
        return sim

    def _init_listeners(self):
        self.logPropagation = False
        self.logApplyingValues = False

    @classmethod
    def build(cls,
              unit: Unit,
              unique_name: str,
              build_dir: Optional[str],
              target_platform=DummyPlatform(),
              do_compile=True) -> "BasicRtlSimulatorVcd":
        """
        Create a hwtSimApi.basic_hdl_simulator based simulation model
        for specified unit and load it to python

        :param unit: interface level unit which you wont prepare for simulation
        :param unique_name: unique name for build directory and python module with simulator
        :param target_platform: target platform for this synthesis
        :param build_dir: directory to store sim model build files,
            if None sim model will be constructed only in memory
        """
        if unique_name is None:
            unique_name = unit._getDefaultName()

        _filter = SerializerFilterDoNotExclude()
        if build_dir is None or not do_compile:
            buff = StringIO()
            store_man = SaveToStream(SimModelSerializer, buff, _filter=_filter)
        else:
            if not os.path.isabs(build_dir):
                build_dir = os.path.join(os.getcwd(), build_dir)
            build_private_dir = os.path.join(build_dir, unique_name)
            store_man = SaveToFilesFlat(SimModelSerializer,
                                        build_private_dir,
                                        _filter=_filter)
            store_man.module_path_prefix = unique_name

        to_rtl(unit,
               name=unique_name,
               target_platform=target_platform,
               store_manager=store_man)

        if build_dir is not None:
            d = build_dir
            dInPath = d in sys.path
            if not dInPath:
                sys.path.insert(0, d)
            if unique_name in sys.modules:
                del sys.modules[unique_name]
            simModule = importlib.import_module(
                unique_name + "." + unique_name,
                package='simModule_' + unique_name)

            if not dInPath:
                sys.path.pop(0)
        else:
            simModule = ModuleType('simModule_' + unique_name)
            # python supports only ~100 opened brackets
            # if exceeded it throws MemoryError: s_push: parser stack overflow
            exec(buff.getvalue(),
                 simModule.__dict__)

        model_cls = simModule.__dict__[unit._name]
        # can not use just function as it would get bounded to class
        return cls(model_cls, unit)

    @internal
    def get_trace_formatter(self, t)\
            -> Tuple[str, int, Callable[[RtlSignalBase, HValue], str]]:
        """
        :return: (vcd type name, vcd width, formatter fn)
        """
        if isinstance(t, (Bits3t, Bits)):
            return (VCD_SIG_TYPE.WIRE, t.bit_length(), VcdBitsFormatter())
        elif isinstance(t, (Enum3t, HEnum)):
            return (VCD_SIG_TYPE.REAL, 1, VcdEnumFormatter())
        else:
            raise ValueError(t)

    def set_trace_file(self, file_name, trace_depth):
        self.create_wave_writer(file_name)
        ww = self.wave_writer
        if ww is not None:
            ww.date(datetime.now())
            ww.timescale(1)

            interface_signals = set()
            self._collect_interface_signals(self.synthesised_unit, self.model, interface_signals)
            self.wave_register_signals(self.synthesised_unit, self.model, None, interface_signals)

            ww.enddefinitions()

    def create_wave_writer(self, file_name):
        self.wave_writer = None

    def finalize(self):
        pass

    def _collect_interface_signals(self,
                                   obj: Union[Interface, Unit],
                                   model: BasicRtlSimModel, res: Set[BasicRtlSimProxy]):
        intfs = getattr(obj, "_interfaces", None)
        if intfs:
            for chIntf in intfs:
                self._collect_interface_signals(chIntf, model, res)
            if isinstance(obj, Unit):
                for u in obj._units:
                    m = getattr(model, u._name + "_inst")
                    if u._shared_component_with is not None:
                        u, _, _ = u._shared_component_with
                    self._collect_interface_signals(u, m, res)
        else:
            sig_name = obj._sigInside.name
            s = getattr(model.io, sig_name)
            res.add(s)

    def wave_register_signals(self,
                              obj: Union[Interface, Unit],
                              model: BasicRtlSimModel,
                              parent: Optional[VcdVarWritingScope],
                              interface_signals: Set[BasicRtlSimProxy]):
        """
        Register signals from interfaces for Interface or :class:`hwt.synthesizer.unit.Unit` instances
        """
        if obj._interfaces:
            if isinstance(obj, Unit):
                name = model._name
            else:
                name = obj._name
            parent_ = self.wave_writer if parent is None else parent

            subScope = parent_.varScope(name)
            self._obj2scope[obj] = subScope

            with subScope:
                # register all subinterfaces
                for chIntf in obj._interfaces:
                    self.wave_register_signals(chIntf, model, subScope, interface_signals)
                if isinstance(obj, Unit):
                    self.wave_register_remaining_signals(subScope, model, interface_signals)
                    # register interfaces from all subunits
                    for u in obj._units:
                        m = getattr(model, u._name + "_inst")
                        if u._shared_component_with is not None:
                            u, _, _ = u._shared_component_with
                        self.wave_register_signals(u, m, subScope, interface_signals)

            return subScope
        else:
            t = obj._dtype
            if isinstance(t, self.supported_type_classes):
                tName, width, formatter = self.get_trace_formatter(t)
                sig_name = obj._sigInside.name
                s = getattr(model.io, sig_name)
                try:
                    parent.addVar(s, sig_name,
                                  tName, width, formatter)
                except VarAlreadyRegistered:
                    pass

    def wave_register_remaining_signals(self, unitScope,
                                        model: BasicRtlSimModel,
                                        interface_signals: Set[BasicRtlSimProxy]):
        for s in model._interfaces:
            if s not in interface_signals and s not in self.wave_writer._idScope:
                t = s._dtype
                if isinstance(t, self.supported_type_classes):
                    tName, width, formatter = self.get_trace_formatter(t)
                    try:
                        unitScope.addVar(s, s._name, tName, width, formatter)
                    except VarAlreadyRegistered:
                        pass

    def logChange(self, nowTime: int,
                  sig: BasicRtlSimProxy,
                  nextVal: HValue,
                  valueUpdater: Union[ValueUpdater, ArrayValueUpdater]):
        """
        This method is called for every value change of any signal.
        """
        pass



from builtins import isinstance
from copy import deepcopy
from typing import Callable, Tuple, Generator, Union, Optional

from hwt.doc_markers import internal
from hwt.hdl.types.array import HArray
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.hdlType import HdlType
from hwt.hdl.types.stream import HStream
from hwt.hdl.types.struct import HStruct, HStructField
from hwt.hdl.types.union import HUnion
from hwt.pyUtils.arrayQuery import iter_with_last
from hwt.synthesizer.typePath import TypePath


def _default_shouldEnterFn(transTmpl: 'TransTmpl') -> Tuple[bool, bool]:
    return (bool(transTmpl.children), not bool(transTmpl.children))


class TransTmpl(object):
    """
    Transaction template for types of constant size

    Container of informations about frames generated from any HType
    (HStruct etc.)
    * contains precalculated address range for all members of type

    :note: Array/Stream items are are stored as a single instance
        so the memory consumption of this object is entirely independent
        on size of arrays which it describes.

    :ivar ~.dtype: type of this item
    :ivar ~.bitAddr: offset of start of this item in bits
    :ivar ~.bitAddrEnd: end of this item in bits
    :ivar ~.parent: object which generated this item, optional TransTmpl
    :ivar ~.origin: object which was template for generating of this item
    :ivar ~.itemCnt: if this transaction template is for array or stream this is
        item count for such an array or stream
    :ivar ~.childrenAreChoice: flag which tells if children are sequence
        or only one of them can be used in same time
    :ivar rel_field_path: path in original data type relative to parent
    """

    def __init__(self, dtype: HdlType, bitAddr: int=0,
                 parent: Optional['TransTmpl']=None,
                 origin: Optional[HStructField]=None,
                 rel_field_path: TypePath=TypePath()):
        self.parent = parent
        assert isinstance(dtype, HdlType), dtype
        assert isinstance(rel_field_path, TypePath), rel_field_path
        assert parent is None or isinstance(parent, TransTmpl), parent
        if origin is None:
            origin = (dtype,)
        else:
            assert isinstance(origin, tuple), origin
        self.origin = origin
        self.dtype = dtype
        self.children = []
        self.itemCnt = None
        self.rel_field_path = rel_field_path
        self._loadFromHType(dtype, bitAddr)

    @internal
    def _loadFromBits(self, dtype: HdlType, bitAddr: int):
        """
        Parse Bits type to this transaction template instance

        :return: address of it's end
        """
        return bitAddr + dtype.bit_length()

    @internal
    def _loadFromHStruct(self, dtype: HdlType, bitAddr: int):
        """
        Parse HStruct type to this transaction template instance

        :return: address of it's end
        """

        for f in dtype.fields:
            t = f.dtype

            isPadding = f.name is None

            if isPadding:
                width = t.bit_length()
                bitAddr += width
            else:
                origin = (*self.origin, f)
                fi = TransTmpl(t, bitAddr,
                               parent=self,
                               origin=origin,
                               rel_field_path=TypePath(f.name,),
                )
                self.children.append(fi)
                bitAddr = fi.bitAddrEnd

        return bitAddr

    @internal
    def _loadFromUnion(self, dtype: HdlType, bitAddr: int) -> int:
        """
        Parse HUnion type to this transaction template instance

        :return: address of it's end
        """
        for f in dtype.fields.values():
            ch = TransTmpl(f.dtype, 0, parent=self,
                           origin=(*self.origin, f),
                           rel_field_path=TypePath(f.name,),
                           )
            self.children.append(ch)
        return bitAddr + dtype.bit_length()

    @internal
    def _loadFromArray(self, dtype: HdlType, bitAddr: int) -> int:
        """
        Parse HArray type to this transaction template instance

        :return: address of it's end
        """
        self.itemCnt = int(dtype.size)
        self.children = TransTmpl(
            dtype.element_t, 0, parent=self,
            origin=(*self.origin, 0),
            rel_field_path=TypePath(0,)
        )
        return bitAddr + self.itemCnt * self.children.bitAddrEnd

    @internal
    def _loadFromHStream(self, dtype: HStream, bitAddr: int) -> int:
        """
        Parse HStream type to this transaction template instance

        :return: address of it's end
        """
        self.children = TransTmpl(
            dtype.element_t, 0, parent=self, origin=self.origin,
            rel_field_path=TypePath(0,))

        if not isinstance(dtype.len_min, int) or dtype.len_min != dtype.len_max:
            raise ValueError("This template is ment only"
                             " for types of constant and finite size")

        self.itemCnt = dtype.len_min
        return bitAddr + dtype.element_t.bit_length() * self.itemCnt

    def _loadFromHType(self, dtype: HdlType, bitAddr: int) -> None:
        """
        Parse any HDL type to this transaction template instance
        """
        self.bitAddr = bitAddr
        childrenAreChoice = False
        if isinstance(dtype, Bits):
            ld = self._loadFromBits
        elif isinstance(dtype, HStruct):
            ld = self._loadFromHStruct
        elif isinstance(dtype, HArray):
            ld = self._loadFromArray
        elif isinstance(dtype, HStream):
            ld = self._loadFromHStream
        elif isinstance(dtype, HUnion):
            ld = self._loadFromUnion
            childrenAreChoice = True
        else:
            raise TypeError("expected instance of HdlType", dtype)

        self.bitAddrEnd = ld(dtype, bitAddr)
        self.childrenAreChoice = childrenAreChoice

    def bit_length(self) -> int:
        """
        :return: number of bits in this transaction
        """
        return self.bitAddrEnd - self.bitAddr

    def walkFlatten(self, offset: int=0,
                    shouldEnterFn=_default_shouldEnterFn
                    ) -> Generator[
            Union[Tuple[Tuple[int, int], 'TransTmpl'], 'OneOfTransaction'],
            None, None]:
        """
        Walk fields in instance of TransTmpl

        :param offset: optional offset for all children in this TransTmpl
        :param shouldEnterFn: function (transTmpl) which returns True
            when field should be split on it's children
        :param shouldEnterFn: function(transTmpl) which should return
            (shouldEnter, shouldUse) where shouldEnter is flag that means
            iterator should look inside of this actual object
            and shouldUse flag means that this field should be used
            (=generator should yield it)
        :return: generator of tuples ((startBitAddress, endBitAddress),
            TransTmpl instance)
        """

        t = self.dtype
        base = self.bitAddr + offset
        end = self.bitAddrEnd + offset

        shouldEnter, shouldYield = shouldEnterFn(self)
        if shouldYield:
            yield ((base, end), self)

        if shouldEnter:
            if isinstance(t, Bits):
                pass
            elif isinstance(t, HStruct):
                for c in self.children:
                    yield from c.walkFlatten(
                        offset,
                        shouldEnterFn)
            elif isinstance(t, (HArray, HStream)):
                itemSize = (self.bitAddrEnd - self.bitAddr) // self.itemCnt
                for i in range(self.itemCnt):
                    if i == 0:
                        c = self.children
                    else:
                        # spot a new array item
                        c = deepcopy(self.children)
                        assert c.rel_field_path == (0,), c.rel_field_path
                        # replace the index
                        c.rel_field_path = TypePath(i, )

                    yield from c.walkFlatten(
                        base + i * itemSize,
                        shouldEnterFn)

            elif isinstance(t, HUnion):
                yield OneOfTransaction(self, offset, shouldEnterFn,
                                       self.children)
            else:
                raise TypeError(t)

    def getFieldPath(self):
        """
        Get field path which specifies the location in original HdlType data type
        """
        path = []
        tmpl = self
        while tmpl is not None:
            path.extend(reversed(tmpl.rel_field_path))
            tmpl = tmpl.parent
        return TypePath(*reversed(path))

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, deepcopy(v, memo))
        c = self.children
        if isinstance(c, TransTmpl):
            c.parent = self
        else:
            for _c in c:
                _c.parent = self

        return result

    def __repr__(self, offset: int=0):
        offsetStr = "".join(["    " for _ in range(offset)])

        try:
            name = self.origin[-1].name
        except (AttributeError, IndexError):
            name = None

        if name:
            name = f" name:{name:s},"
        else:
            name = ""

        s = f"{offsetStr:s}<TransTmpl{name} start:{self.bitAddr:d}, end:{self.bitAddrEnd:d}"
        if isinstance(self.dtype, (HArray, HStream)):
            s_buff = [
                s, f", itemCnt:{self.itemCnt:d}\n",
                self.children.__repr__(offset=offset + 1), "\n",
                offsetStr, ">"
            ]
            return "".join(s_buff)
        elif not self.children:
            return s + ">"

        buff = [s, ]
        for isLast, child in iter_with_last(self.children):
            buff.append(child.__repr__(offset=offset + 1))
            if self.childrenAreChoice and not isLast:
                buff.append(offsetStr + "    <OR>")

        buff.append(offsetStr + ">")
        return "\n".join(buff)


class OneOfTransaction(object):
    """
    Container of possible transactions for transactions deriverd
    from HUnion type

    :ivar ~.parent: parent TransTmpl instance
    :ivar ~.offset: bit addr offset in parent type structure
    :ivar ~.shouldEnterFn: function(transTmpl) which should
        return (shouldEnter, shouldUse)
        where shouldEnter is flag that means iterator should look
        inside of this actual object and shouldUse flag means that this field
        should be used (=generator should yield it)
    :ivar ~.possibleTransactions: tuple of TransTmpl instances
        from which only one can be used in same time
    """

    def __init__(self, parent: TransTmpl,
                 offset: int,
                 shouldEnterFn: Callable[[TransTmpl], Tuple[bool, bool]],
                 possibleTransactions: Tuple[TransTmpl]):
        self.parent = parent
        self.offset = offset
        self.shouldEnterFn = shouldEnterFn
        self.possibleTransactions = possibleTransactions

    def walkFlattenChilds(self) -> Generator[
            Union[Tuple[Tuple[int, int], TransTmpl], 'OneOfTransaction'],
            None, None]:
        """
        :return: generator of generators of tuples
            ((startBitAddress, endBitAddress), TransTmpl instance)
            for each possiblility in this transaction
        """
        for p in self.possibleTransactions:
            yield p.walkFlatten(offset=self.offset,
                                shouldEnterFn=self.shouldEnterFn)



# -*- coding: utf-8 -*-

from io import StringIO

from hwt.serializer.serializer_config import DummySerializerConfig
from hwt.serializer.serializer_filter import SerializerFilterDoNotExclude
from hwt.serializer.store_manager import SaveToStream, StoreManager
from hwt.serializer.vhdl import Vhdl2008Serializer
from hwt.synthesizer.dummyPlatform import DummyPlatform
from hwt.synthesizer.unit import Unit, HdlConstraintList
from hwt.constraints import _get_absolute_path


def to_rtl(unit_or_cls: Unit, store_manager: StoreManager,
           name: str=None,
           target_platform=DummyPlatform()):
    """
    Convert unit to RTL using specified serializer

    :param unitOrCls: unit instance or class, which should be converted
    :param name: name override of top unit (if is None name is derived
        form class name)
    :param target_platform: meta-informations about target platform, distributed
        on every unit under _target_platform attribute
        before Unit._impl() is called
    """
    if isinstance(unit_or_cls, Unit):
        u = unit_or_cls
    else:
        u = unit_or_cls()

    u._target_platform = target_platform
    u._store_manager = store_manager
    u._loadDeclarations()
    if name is not None:
        assert isinstance(name, str)
        u._hdl_module_name = u._name = name

    # serialize all unit instances to HDL code
    constraints = HdlConstraintList()
    for _, obj in u._to_rtl(target_platform, store_manager):
        # collect constraints directly in current component
        constraints.extend(obj._constraints)

        if obj._shared_component_with:
            # if the instance is shared with something else make
            # the paths in constraints relative to a component
            path_old = _get_absolute_path(obj._shared_component_with[0])
            path_new = _get_absolute_path(obj)

            for c in _Unit_constraints_copy_recursively(
                    obj, path_old, path_new):
                constraints.append(c)

    if constraints:
        # serialize all constraints in design
        store_manager.write(constraints)

    return store_manager


def _Unit_constraints_copy_recursively(u: Unit, path_orig: Unit, path_new: Unit):
    if u._shared_component_with:
        assert not u._constraints
        assert not u._units
        orig_u, _, _ = u._shared_component_with
        yield from _Unit_constraints_copy_recursively(
            orig_u, (*path_orig[:-1], orig_u), path_new)
    else:
        for c in u._constraints:
            yield c._copy_with_root_upadate(path_orig, path_new)

        for su in u._units:
            yield from _Unit_constraints_copy_recursively(
                su, (*path_orig, su), (*path_new, su))


def to_rtl_str(unit_or_cls: Unit,
               serializer_cls=Vhdl2008Serializer, name: str=None,
               target_platform=DummyPlatform()):
    """
    Generate HDL string and return it
    """
    buff = StringIO()
    store_manager = SaveToStream(serializer_cls, buff)
    to_rtl(unit_or_cls, store_manager, name, target_platform)
    return buff.getvalue()


def serializeAsIpcore(unit, folderName=".", name=None,
                      serializer_cls=Vhdl2008Serializer,
                      target_platform=DummyPlatform()):
    """
    Create an IPCore package
    """
    from hwt.serializer.ip_packager import IpPackager
    p = IpPackager(unit, name=name,
                   serializer_cls=serializer_cls,
                   target_platform=target_platform)
    p.createPackage(folderName)
    return p


def synthesised(u: Unit, target_platform=DummyPlatform()):
    """
    Elaborate design without producing any HDL
    """
    sm = StoreManager(DummySerializerConfig,
                      _filter=SerializerFilterDoNotExclude())
    if not hasattr(u, "_interfaces"):
        u._loadDeclarations()

    for _ in u._to_rtl(target_platform, sm):
        pass
    return u



from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import AllOps
from hwt.hdl.statements.ifContainter import IfContainer
from hwt.hdl.types.hdlType import HdlType
from hwt.synthesizer.rtlLevel.constants import NOT_SPECIFIED
from hwt.synthesizer.rtlLevel.mainBases import RtlMemoryBase
from hwt.synthesizer.rtlLevel.rtlSignal import RtlSignal


class RtlSyncSignal(RtlMemoryBase, RtlSignal):
    """
    Syntax sugar,
    every write is made to next signal, "next" is assigned
    to main signal on every clock rising edge
    """

    def __init__(self, ctx: 'RtlNetlist', name: str, var_type: HdlType, def_val=None, nop_val=NOT_SPECIFIED):
        """
        :param ~.ctx: context in which is sig. created (instance of RtlNetlist)
        :param ~.name: suggested name
        :param ~.var_type: type of signal
        :param ~.def_val: default value for signal
            (used as def. val in hdl and for reset)
        """
        super().__init__(ctx, name, var_type, def_val)
        if nop_val is NOT_SPECIFIED:
            nop_val = self
        self.next = RtlSignal(ctx, name + "_next", var_type,
                              nop_val=nop_val)

    def _getDestinationSignalForAssignmentToThis(self):
        """
        :see: :func:` hwt.synthesizer.rtlLevel.rtlSignal.RtlSignal`
        """
        return self.next

    def _getAssociatedClk(self):
        d = self.singleDriver()
        assert isinstance(d, IfContainer), d
        cond = d.cond.singleDriver()
        assert isinstance(cond, Operator) and cond.operator is AllOps.RISING_EDGE, cond
        return cond.operands[0]

    def _getAssociatedRst(self):
        d = self.singleDriver()
        assert isinstance(d, IfContainer), d
        cond = d.cond.singleDriver()
        assert isinstance(cond, Operator) and cond.operator is AllOps.RISING_EDGE, cond
        assert len(d.ifTrue) == 1
        reset_if = d.ifTrue[0]
        return reset_if.cond



from hwt.hdl.sensitivityCtx import SensitivityCtx
from hwt.doc_markers import internal


class HValue():
    """
    Wrap around hdl value with overloaded operators
    http://www.rafekettler.com/magicmethods.html

    operators are overloaded in every type separately
    """
    __slots__ = ["_dtype", "val", "vld_mask"]

    def __init__(self, dtype, val, vld_mask):
        """
        :param val: pythonic value representing this value
        :param dtype: data type object from which this value was derived from
        :param vld_mask: validity mask for value
        """
        self._dtype = dtype
        self.val = val
        self.vld_mask = vld_mask

    def _is_full_valid(self):
        return self.vld_mask == self._dtype.all_mask()

    def _auto_cast(self, toT):
        "type cast to compatible type"
        return self._dtype.auto_cast(self, toT)

    def _reinterpret_cast(self, toT):
        "type cast to type of same size"
        return self._dtype.reinterpret_cast(self, toT)

    def staticEval(self):
        return self.__copy__()

    def __copy__(self):
        return self.__class__(self._dtype, self.val, self.vld_mask)

    @internal
    def __hash__(self):
        return hash((self._dtype, self.val, self.vld_mask))

    def __repr__(self):
        if self._is_full_valid():
            vld_mask = ""
        else:
            vld_mask = ", mask {0:x}".format(self.vld_mask)
        return "<{0:s} {1:s}{2:s}>".format(
            self.__class__.__name__, repr(self.val), vld_mask)

    @classmethod
    def _from_py(cls, typeObj, val, vld_mask):
        """
        from_py without value normalization and type checking
        """
        return cls(typeObj, val, vld_mask)

    @classmethod
    def from_py(cls, typeObj, val, vld_mask=None):
        raise NotImplementedError(
            f"from_py fn is not implemented for {cls}")

    def __int__(self):
        if isinstance(self, HValue) or self._const:
            if self._is_full_valid():
                return int(self.val)
            else:
                raise ValueError(f"HValue of {self} is not fully defined")

        raise ValueError(
            f"HValue of {self} is not constant it can be statically solved")

    def __bool__(self):
        if isinstance(self, HValue) or self._const:
            if self._is_full_valid():
                return bool(self.val)
            else:
                raise ValueError(f"HValue of {self} is not fully defined")

        raise ValueError(
            f"HValue of {self} is not constant it can be statically solved")

    def __eq__(self, other):
        if areHValues(self, other):
            return self._dtype == other._dtype and \
                self.vld_mask == other.vld_mask and\
                self.val == other.val
        else:
            return super().__eq__(other)

    def _eq(self, other):
        raise TypeError()

    def __ne__(self, other):
        eq = self._eq(other)
        eq.val = not eq.val
        return eq

    def _walk_sensitivity(self, casualSensitivity: set, seen: set, ctx: SensitivityCtx):
        seen.add(self)

    # def __pos__(self):
    #     raise TypeError()
    #
    # def __neg__(self):
    #     raise TypeError()
    #
    # def __abs__(self):
    #     raise TypeError()
    #
    # def __invert__(self):
    #     raise TypeError()
    #
    # def __round__(self, n):
    #     raise TypeError()
    #
    # def __floor__(self):
    #     raise TypeError()
    #
    # def __ceil__(self):
    #     raise TypeError()
    #
    # def __add__(self, other):
    #     raise TypeError()
    #
    # def __sub__(self, other):
    #     raise TypeError()
    #
    # def __mul__(self, other):
    #     raise TypeError()
    #
    # def __floordiv__(self, other):
    #     raise TypeError()
    #
    # def __div__(self, other):
    #     raise TypeError()
    #
    # def __truediv__(self, other):
    #     raise TypeError()
    #
    # def __mod__(self, other):
    #     raise TypeError()
    #
    # def __divmod__(self, other):
    #     raise TypeError()
    #
    # def __pow__(self, other):
    #     raise TypeError()
    #
    # def __lshift__(self, other):
    #     raise TypeError()
    #
    # def __rshift__(self, other):
    #     raise TypeError()
    #
    # def __and__(self, other):
    #     raise TypeError()
    #
    # def __or__(self, other):
    #     raise TypeError()
    #
    # def __xor__(self, other):
    #     raise TypeError()
    #
    # def _concat(self, other):
    #     raise TypeError()
    #
    # def __lt__(self, other):
    #     raise TypeError()
    #
    # def __le__(self, other):
    #     raise TypeError()
    #
    # def __gt__(self, other):
    #     raise TypeError()
    #
    # def __ge__(self, other):
    #     raise TypeError()
    #
    # def _onRisingEdge(self):
    #     raise TypeError()
    #
    # def _onFallingEdge(self):
    #     raise TypeError()


def areHValues(*items):
    """
    :return: True if all arguments are instances of HValue class else False
    """
    for i in items:
        if not isinstance(i, HValue):
            return False
    return True



from copy import copy
from typing import Union

from hdlConvertorAst.hdlAst._defs import HdlIdDef
from hdlConvertorAst.hdlAst._expr import HdlValueInt, HdlOp, HdlOpType, \
    HdlValueId
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model.utils import hdl_getattr, \
    hdl_call
from hdlConvertorAst.translate.common.name_scope import ObjectForNameNotFound
from hwt.hdl.types.arrayVal import HArrayVal
from hwt.hdl.types.bitsVal import BitsVal
from hwt.hdl.types.defs import SLICE
from hwt.hdl.types.enum import HEnum
from hwt.hdl.types.enumVal import HEnumVal
from hwt.hdl.types.sliceVal import HSliceVal
from hwt.hdl.value import HValue
from hwt.hdl.variables import SignalItem
from hwt.serializer.generic.value import ToHdlAst_Value
from hwt.serializer.simModel.value import ToHdlAstSimModel_value


class ToHdlAstHwt_value(ToHdlAst_Value):
    NONE = HdlValueId("None")
    SLICE = HdlValueId("SLICE", obj=SLICE)

    def is_suitable_for_const_extract(self, val: HValue):
        # full valid values can be represented as int and do not have any
        # constructor overhead, entirely invalid values can be represented by None
        return not val._is_full_valid() and not isinstance(val._dtype, HEnum)

    def as_hdl_BitsVal(self, val: BitsVal):
        isFullVld = val._is_full_valid()
        if not self._valueWidthRequired:
            if isFullVld:
                return HdlValueInt(val.val, None, 16)
            elif val.vld_mask == 0:
                return self.NONE

        t = self.as_hdl_HdlType_bits(val._dtype, declaration=False)
        c = hdl_getattr(t, "from_py")
        args = [HdlValueInt(val.val, None, 16), ]
        if not isFullVld:
            args.append(HdlValueInt(val.vld_mask, None, 16))

        return hdl_call(c, args)

    def as_hdl_SignalItem(self, si: Union[SignalItem, HdlIdDef], declaration=False):
        if declaration:
            if isinstance(si, HdlIdDef):
                new_si = copy(si)
                new_si.type = self.as_hdl_HdlType(si.type)
                if si.value is not None:
                    new_si.value = self.as_hdl_Value(si.value)
                return new_si
            else:
                raise NotImplementedError()
        else:
            # if isinstance(si, SignalItem) and si._const:
            #    # to allow const cache to extract constants
            #    return self.as_hdl_Value(si._val)
            if si.hidden and si.origin is not None:
                return self.as_hdl(si.origin)
            else:
                return HdlValueId(si.name, obj=si)

    def as_hdl_DictVal(self, val):
        return ToHdlAstSimModel_value.as_hdl_DictVal(self, val)

    def as_hdl_HArrayVal(self, val: HArrayVal):
        if not val.vld_mask:
            return self.NONE
        # else:
        #    if len(val.val) == val._dtype.size:
        #        allValuesSame = True
        #        values = iter(val.val.values())
        #        reference = next(values)
        #         for v in values:
        #             if allValuesSame:
        #                 allValuesSame = isSameHVal(reference, v)
        #             else:
        #                 break
        #        if allValuesSame:
        #            # all values of items in array are same, use generator
        #            # exression
        #            raise NotImplementedError()
        #            return "[%s for _ in range(%d)]" % (self.Value(reference))

        # if value can not be simplified it is required to serialize it item
        # by item
        return self.as_hdl_DictVal(val.val)

    def as_hdl_HSliceVal(self, val: HSliceVal):
        if val._is_full_valid():
            return HdlOp(
                HdlOpType.DOWNTO, [
                    HdlValueInt(int(val.val.start), None, None),
                    HdlValueInt(int(val.val.stop), None, None)
                ])
        else:
            raise NotImplementedError()
            return "HSliceVal(slice(%s, %s, %s), SLICE, %d)" % (
                self.as_hdl_Value(val.val.start),
                self.as_hdl_Value(val.val.stop),
                self.as_hdl_Value(val.val.step),
                val.vld_mask)

    def as_hdl_HEnumVal(self, val: HEnumVal):
        try:
            t_name = self.name_scope.get_object_name(val._dtype)
        except ObjectForNameNotFound:
            if self.debug:
                t_name = val._dtype.name
            else:
                raise

        if val.vld_mask:
            try:
                name = self.name_scope.get_object_name(val)
            except ObjectForNameNotFound:
                if self.debug:
                    name = val.val
                else:
                    raise

            return hdl_getattr(HdlValueId(t_name, obj=val._dtype), name)
        else:
            return hdl_call(hdl_getattr(HdlValueId(t_name, obj=val._dtype), "from_py"),
                            [None, ])



from builtins import isinstance
from typing import Union

from hdlConvertorAst.hdlAst import HdlValueInt
from hdlConvertorAst.hdlAst._expr import HdlValueId, HdlOpType, HdlOp
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model.utils import hdl_call
from hdlConvertorAst.translate.common.name_scope import LanguageKeyword
from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import AllOps
from hwt.hdl.types.defs import BIT, INT
from hwt.hdl.value import HValue
from hwt.serializer.exceptions import UnsupportedEventOpErr
from hwt.serializer.generic.ops import HWT_TO_HDLCONVEROTR_OPS
from hwt.synthesizer.rtlLevel.rtlSignal import RtlSignal


zero, one = BIT.from_py(0), BIT.from_py(1)


class ToHdlAstVerilog_ops():
    SIGNED = HdlValueId("$signed", obj=LanguageKeyword())
    UNSIGNED = HdlValueId("$unsigned", obj=LanguageKeyword())
    op_transl_dict = {
        **HWT_TO_HDLCONVEROTR_OPS,
        AllOps.INDEX: HdlOpType.INDEX,
    }

    def _operandIsAnotherOperand(self, operand):
        if isinstance(operand, RtlSignal) and operand.hidden\
                and isinstance(operand.origin, Operator):
            return True

    def as_hdl_operand(self, operand: Union[RtlSignal, HValue], i: int,
                       operator: Operator):

        # [TODO] if operand is concatenation and parent operator
        #        is not concatenation operand should be extracted
        #        as tmp variable
        #        * maybe flatten the concatenations
        if operator.operator != AllOps.CONCAT\
                and self._operandIsAnotherOperand(operand)\
                and operand.origin.operator == AllOps.CONCAT:
            _, tmpVar = self.tmpVars.create_var_cached("tmp_concat_", operand._dtype, def_val=operand)
            # HdlAssignmentContainer(tmpVar, operand, virtual_only=True)
            operand = tmpVar
        elif operator.operator == AllOps.INDEX and i == 0 and self._operandIsAnotherOperand(operand):
            _, tmpVar = self.tmpVars.create_var_cached("tmp_index_", operand._dtype, def_val=operand)
            operand = tmpVar

        oper = operator.operator
        width = None
        if not isinstance(operand, RtlSignal) and operand._dtype == INT and\
           oper not in [AllOps.BitsAsUnsigned,
                        AllOps.BitsAsVec,
                        AllOps.BitsAsSigned,
                        AllOps.INDEX] and\
                operator.result is not None and\
                not operator.result._dtype == INT:
            # have to lock the width
            for o in operator.operands:
                try:
                    bl = o._dtype.bit_length
                except AttributeError:
                    bl = None
                if bl is not None:
                    width = bl()
                    break

            assert width is not None, (operator, operand)
        hdl_op = self.as_hdl_Value(operand)
        if width is not None:
            if isinstance(hdl_op, HdlValueInt):
                assert isinstance(width, int), width
                hdl_op.bits = width
            else:
                return HdlOp(HdlOpType.APOSTROPHE, [self.as_hdl_int(width), hdl_op])
        return hdl_op

    def as_hdl_Operator(self, op: Operator):
        ops = op.operands
        o = op.operator

        if o == AllOps.TERNARY:
            if ops[1] == one and ops[2] == zero:
                # ignore redundant x ? 1 : 0
                return self.as_hdl_cond(ops[0], True)
            else:
                op0 = self.as_hdl_cond(ops[0], True)
                op1 = self.as_hdl_operand(ops[1], 1, op)
                op2 = self.as_hdl_operand(ops[2], 2, op)
                return HdlOp(HdlOpType.TERNARY, [op0, op1, op2])
        elif o == AllOps.RISING_EDGE or o == AllOps.FALLING_EDGE:
            raise UnsupportedEventOpErr()
        elif o in [AllOps.BitsAsUnsigned, AllOps.BitsAsVec, AllOps.BitsAsSigned]:
            op0, = ops
            do_cast = bool(op0._dtype.signed) != bool(op.result._dtype.signed)

            op_hdl = self.as_hdl_operand(op0, 0, op)
            if do_cast:
                if bool(op0._dtype.signed):
                    cast = self.SIGNED
                else:
                    cast = self.UNSIGNED
                return hdl_call(cast, [op_hdl, ])
            else:
                return op_hdl
        else:
            _o = self.op_transl_dict[o]
            return HdlOp(_o, [self.as_hdl_operand(o2, i, op)
                              for i, o2 in enumerate(ops)])



from itertools import chain

from hdlConvertorAst.hdlAst._defs import HdlIdDef
from hdlConvertorAst.hdlAst._structural import HdlModuleDef, \
    HdlCompInst
from hwt.doc_markers import internal
from hwt.hdl.operator import Operator, isConst
from hwt.hdl.operatorDefs import AllOps
from hwt.hdl.statements.assignmentContainer import HdlAssignmentContainer
from hwt.hdl.statements.codeBlockContainer import HdlStmCodeBlockContainer
from hwt.hdl.statements.statement import HdlStatement
from hwt.hdl.statements.switchContainer import SwitchContainer
from hwt.hdl.types.array import HArray
from hwt.hdl.value import HValue
from hwt.serializer.resourceAnalyzer.utils import ResourceContext
from hwt.synthesizer.rtlLevel.mark_visibility_of_signals_and_check_drivers import walk_assignments
from hwt.synthesizer.rtlLevel.rtlSignal import RtlSignal
from hwt.synthesizer.unit import Unit


@internal
def _count_mux_inputs_for_outputs(stm: HdlStatement, cnt):
    if isinstance(stm, HdlAssignmentContainer):
        cnt[stm.dst] += 1
    else:
        for _stm in stm._iter_stms():
            if isinstance(_stm, HdlAssignmentContainer):
                cnt[_stm.dst] += 1
            else:
                _count_mux_inputs_for_outputs(_stm, cnt)


@internal
def count_mux_inputs_for_outputs(stm):
    cnt = {o: 0 for o in stm._outputs}
    _count_mux_inputs_for_outputs(stm, cnt)
    return cnt


# operators which do not consume a hw resorce directly
IGNORED_OPERATORS = {
    AllOps.BitsAsSigned,
    AllOps.BitsAsUnsigned,
    AllOps.BitsAsVec,
    AllOps.RISING_EDGE,
    AllOps.FALLING_EDGE,
}


class ResourceAnalyzer():
    """
    Serializer which does not products any output just collect informations
    about used resources

    :attention: Use instance of ResourceAnalyzer instead of class
    """
    _keywords_dict = {}

    def __init__(self):
        self.context = ResourceContext(None)

    @internal
    def visit_HdlStmCodeBlockContainer_operators(self, sig: RtlSignal, synchronous):
        ctx = self.context
        seen = ctx.seen
        for d in sig.drivers:
            if (not isinstance(d, Operator)
                    or d in seen):
                continue

            skip_op = d.operator in IGNORED_OPERATORS
            if not skip_op:
                if d.operator == AllOps.EQ:
                    o1 = d.operands[1]
                    if (isinstance(o1, HValue)
                            and o1._dtype.bit_length() == 1
                            and o1.val):
                        # to bool conversion
                        skip_op = True
                elif d.operator == AllOps.INDEX:
                    o1 = d.operands[1]
                    skip_op = True
                    if isConst(o1):
                        # constant signal slice
                        pass
                    else:
                        o0 = d.operands[0]
                        if isinstance(o0._dtype, HArray):
                            ctx.registerRAM_read_port(o0, o1, synchronous)
                        else:
                            ctx.registerMUX(d, sig, 2)
                elif d.operator == AllOps.TERNARY:
                    o1 = d.operands[1]
                    o2 = d.operands[2]
                    if (isConst(o1)
                            and bool(o1)
                            and isConst(o2)
                            and not bool(o2)):
                        # to bit conversion
                        skip_op = True
                    else:
                        raise NotImplementedError("Ternary as mux")

            if not skip_op:
                ctx.registerOperator(d)

            for op in d.operands:
                if (not isinstance(op, RtlSignal)
                        or not op.hidden
                        or op in seen):
                    continue
                self.visit_HdlStmCodeBlockContainer_operators(op, synchronous)

    def visit_HdlStmCodeBlockContainer(self, proc: HdlStmCodeBlockContainer) -> None:
        """
        Guess resource usage from HdlStmCodeBlockContainer
        """
        ctx = self.context
        seen = ctx.seen
        for stm in proc.statements:
            encl = stm._enclosed_for
            full_ev_dep = stm._event_dependent_from_branch == 0
            now_ev_dep = stm._event_dependent_from_branch is not None
            ev_dep = full_ev_dep or now_ev_dep

            out_mux_dim = count_mux_inputs_for_outputs(stm)
            for o in stm._outputs:
                if o in seen:
                    continue

                i = out_mux_dim[o]
                if isinstance(o._dtype, HArray):
                    assert i == 1, (o, i, " only one ram port per HdlStmCodeBlockContainer")
                    for a in walk_assignments(stm, o):
                        assert len(a.indexes) == 1, ("has to have single address per RAM port", a.indexes)
                        addr = a.indexes[0]
                    ctx.registerRAM_write_port(o, addr, ev_dep)
                elif ev_dep:
                    ctx.registerFF(o)
                    if i > 1:
                        ctx.registerMUX(stm, o, i)
                elif o not in encl:
                    ctx.registerLatch(o)
                    if i > 1:
                        ctx.registerMUX(stm, o, i)
                elif i > 1:
                    ctx.registerMUX(stm, o, i)
                else:
                    # just a connection
                    continue

            if isinstance(stm, SwitchContainer):
                caseEqs = set([stm.switchOn._eq(c[0]) for c in stm.cases])
                inputs = chain(
                    [sig for sig in stm._inputs if sig not in caseEqs], [stm.switchOn])
            else:
                inputs = stm._inputs

            for i in inputs:
                # discover only internal signals in this statements for
                # operators
                if not i.hidden or i in seen:
                    continue

                self.visit_HdlStmCodeBlockContainer_operators(i, ev_dep)

    def visit_HdlModuleDef(self, m: HdlModuleDef) -> None:
        for o in m.objs:
            if isinstance(o, HdlStmCodeBlockContainer):
                self.visit_HdlStmCodeBlockContainer(o)
            elif isinstance(o, HdlCompInst):
                self.visit_HdlCompInst(o)
            else:
                assert isinstance(o, HdlIdDef), o

    def visit_HdlCompInst(self, o: HdlCompInst) -> None:
        raise NotImplementedError()

    # [TODO] constant to ROMs
    def visit_Unit(self, u: Unit):
        self.visit_HdlModuleDef(u._ctx.arch)

    def report(self):
        ctx = self.context
        ctx.finalize()
        return ctx.resources



from typing import Generator

from hwt.doc_markers import internal
from hwt.hdl.portItem import HdlPortItem
from hwt.hdl.sensitivityCtx import SensitivityCtx
from hwt.hdl.statements.statement import HdlStatement
from hwt.hdl.types.hdlType import HdlType
from hwt.hdl.value import HValue
from hwt.hdl.variables import SignalItem
from hwt.pyUtils.uniqList import UniqList
from hwt.synthesizer.rtlLevel.constants import NOT_SPECIFIED
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.synthesizer.rtlLevel.signalUtils.exceptions import SignalDriverErr, \
    SignalDriverErrType
from hwt.synthesizer.rtlLevel.signalUtils.ops import RtlSignalOps


class RtlSignal(RtlSignalBase, SignalItem, RtlSignalOps):
    """
    RtlSignal signal is container of connection
    between statements and operators

    :ivar ~.endpoints: UniqList of operators and statements
        for which this signal is driver.
    :ivar ~.drivers: UniqList of operators and statements
        which can drive this signal.
        If driver is statemet tree only top statement is present.
    :ivar ~._usedOps: A dictionary of used operators which can be reused.
    :ivar ~._usedOpsAlias: A dictionary tuple of operator and operands to set of tules of operator and operands,
        used to resolve which combination of the operator and operands resulted in to same result.
    :ivar ~.hiden: means that this signal is part of expression
        and should not be rendered
    :ivar ~._nop_val: value which is used to fill up statements when no other
            value is assigned, use NOT_SPECIFIED to disable
    :ivar ~._const: flag which tell that this signal can not have any other driver
        than a default value

    :cvar __instCntr: counter used for generating instance ids
    :ivar ~._instId: internally used only for intuitive sorting of statements
        in serialized code
    """
    __instCntr = 0

    def __init__(self, ctx: 'RtlNetlist', name: str, dtype: HdlType, def_val=None, nop_val=NOT_SPECIFIED,
                 virtual_only=False, is_const=False):
        """
        :param ctx: context - RtlNetlist which is this signal part of
        :param name: name hint for this signal, if is None name
            is chosen automatically
        :param def_val: value which is used for reset and as default value
            in hdl
        :param nop_val: value which is used to fill up statements when no other
            value is assigned, use NOT_SPECIFIED to dissable
        :param is_const: flag which tell that this signal can not have any other driver
            than a default value
        """

        if name is None:
            name = "sig_"
            self.hasGenericName = True
        else:
            self.hasGenericName = False

        assert isinstance(dtype, HdlType)
        super(RtlSignal, self).__init__(name, dtype, def_val, virtual_only=virtual_only)
        self.ctx = ctx

        if ctx:
            # params does not have any context on created
            # and it is assigned after param is bounded to unit or interface
            ctx.signals.add(self)

        # set can not be used because hash of items are changing
        self.endpoints = UniqList()
        self.drivers = UniqList()
        self._usedOps = {}
        self._usedOpsAlias = {}
        self.hidden = True
        self._instId = RtlSignal._nextInstId()

        self._nop_val = nop_val
        self._const = is_const
        self.origin = None

    @internal
    @classmethod
    def _nextInstId(cls):
        """
        Get next instance id
        """
        i = cls.__instCntr
        cls.__instCntr += 1
        return i

    def staticEval(self):
        # operator writes in self._val new value
        driven_by_def_val = True
        if self.drivers:
            for d in self.drivers:
                if isinstance(d, HdlPortItem):
                    assert d.getInternSig() is self, (d, self)
                    continue
                d.staticEval()
                driven_by_def_val = False

        if driven_by_def_val:
            if isinstance(self.def_val, RtlSignal):
                self._val = self.def_val._val.staticEval()
            else:
                # _val is invalid initialization value
                self._val = self.def_val.__copy__()

        if not isinstance(self._val, HValue):
            raise ValueError(
                "Evaluation of signal returned not supported object (%r)"
                % (self._val,))

        return self._val

    def singleDriver(self):
        """
        Returns a first driver if signal has only one driver.
        """
        d_cnt = len(self.drivers)
        if d_cnt == 0:
            raise SignalDriverErr([(SignalDriverErrType.MISSING_DRIVER, self), ])
        elif d_cnt > 1:
            raise SignalDriverErr([(SignalDriverErrType.MULTIPLE_COMB_DRIVERS, self), ])

        return self.drivers[0]

    @internal
    def _walk_sensitivity(self, casualSensitivity: set, seen: set, ctx: SensitivityCtx):
        seen.add(self)

        if self._const:
            return

        if not self.hidden:
            casualSensitivity.add(self)
            return

        try:
            op = self.singleDriver()
        except SignalDriverErr:
            op = None

        if op is None or isinstance(op, HdlStatement):
            casualSensitivity.add(self)
            return

        op._walk_sensitivity(casualSensitivity, seen, ctx)

    @internal
    def _walk_public_drivers(self, seen: set) -> Generator["RtlSignal", None, None]:
        """
        Walk all non hiden signals in an expression
        """
        seen.add(self)
        if not self.hidden:
            yield self
            return

        assert self.drivers, self

        for d in self.drivers:
            # d has to be operator otherwise this signal would be public itself
            assert not isinstance(d, HdlStatement), (d.__class__)
            yield from d._walk_public_drivers(seen)



from hwt.doc_markers import internal
from hwt.hdl.constants import DIRECTION
from hwt.hdl.sensitivityCtx import SensitivityCtx
from hwt.hdl.statements.statement import HwtSyntaxError


class HdlPortItem():
    """
    HDL entity/module/component port item
    Used to split signal paths on compomponent boundary.

    :note: src/dst are named based on input output signal direction
        bouth dst and src can be parent/component signal, it depends on direction
    """

    def __init__(self, name, direction, dtype, unit):
        self.name = name
        self.unit = unit
        self._dtype = dtype
        self.direction = direction
        self.src = None
        self.dst = None

    @internal
    def connectOuterSig(self, signal):
        """
        Connect to port item on subunit
        """
        if self.direction == DIRECTION.IN:
            if self.src is not None:
                raise HwtSyntaxError(
                    "Port %s is already associated with %r"
                    % (self.name, self.src))
            self.src = signal
            signal.endpoints.append(self)

        elif self.direction == DIRECTION.OUT:
            if self.dst is not None:
                raise HwtSyntaxError(
                    "Port %s is already associated with %r"
                    % (self.name, self.dst))
            self.dst = signal
            signal.drivers.append(self)

        else:
            raise NotImplementedError(self)

        signal.hidden = False
        signal.ctx.subUnits.add(self.unit)

    @internal
    def connectInternSig(self, signal):
        """
        Connect signal from internal side of of this component to this port.
        """
        if self.direction == DIRECTION.OUT:
            if self.src is not None:
                raise HwtSyntaxError(
                    "Port %s is already associated with signal %s"
                    % (self.name, str(self.src)))
            self.src = signal
            self.src.endpoints.append(self)

        elif self.direction == DIRECTION.IN:
            if self.dst is not None:
                raise HwtSyntaxError(
                    "Port %s is already associated with signal %s"
                    % (self.name, str(self.dst)))
            self.dst = signal
            self.dst.drivers.append(self)
        else:
            raise NotImplementedError(self.direction)

    @internal
    def getInternSig(self):
        """
        return signal inside unit which has this port
        """
        d = self.direction
        if d == DIRECTION.IN:
            return self.dst
        elif d == DIRECTION.OUT:
            return self.src
        else:
            raise NotImplementedError(d)

    @internal
    def getOuterSig(self):
        """
        return signal inside unit which has this port
        """
        d = self.direction
        if d == DIRECTION.OUT:
            return self.dst
        elif d == DIRECTION.IN:
            return self.src
        else:
            raise NotImplementedError(d)

    @internal
    def _walk_sensitivity(self, casualSensitivity: set, seen: set, ctx: SensitivityCtx):
        return
        yield

    def __repr__(self):
        return f"<{self.__class__.__name__:s} src:{self.src}, dst:{self.dst}>"



from hwt.doc_markers import internal
from hwt.hdl.operator import Operator
from hwt.hdl.portItem import HdlPortItem
from hwt.hdl.statements.assignmentContainer import HdlAssignmentContainer
from hwt.hdl.statements.codeBlockContainer import HdlStmCodeBlockContainer
from hwt.hdl.statements.ifContainter import IfContainer
from hwt.hdl.statements.statement import HdlStatement
from hwt.hdl.statements.switchContainer import SwitchContainer
from hwt.synthesizer.interfaceLevel.interfaceUtils.utils import walkPhysInterfaces
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from ipCorePackager.constants import DIRECTION


@internal
def walkInputsForSpecificOutput(output_sig: RtlSignalBase, stm: HdlStatement):
    if output_sig not in stm._outputs:
        return
    if isinstance(stm, HdlAssignmentContainer):
        assert stm.dst is output_sig
        yield from stm._inputs
    elif isinstance(stm, IfContainer):
        yield stm.cond
        for c, _ in stm.elIfs:
            yield c
        for _stm in stm._iter_stms():
            yield from walkInputsForSpecificOutput(output_sig, _stm)
    elif isinstance(stm, SwitchContainer):
        yield stm.switchOn
        for _stm in stm._iter_stms():
            yield from walkInputsForSpecificOutput(output_sig, _stm)
    elif isinstance(stm, HdlStmCodeBlockContainer):
        for _stm in stm._iter_stms():
            yield from walkInputsForSpecificOutput(output_sig, _stm)
    else:
        raise NotImplementedError(stm)


@internal
def removeUnconnectedSignals(netlist: "RtlNetlist"):
    """
    Remove signal if does not affect output

    :attention: does not remove signals in cycles which does not affect outputs
    """
    toSearch = [s for s, d in netlist.interfaces.items() if d != DIRECTION.IN]
    seen = set(toSearch)
    for c in netlist.subUnits:
        toSearch.extend(sig._sig for sig in walkPhysInterfaces(c))

    while toSearch:
        _toSearch = []
        for sig in toSearch:
            for e in sig.drivers:
                if isinstance(e, Operator):
                    inputs = e.operands
                elif isinstance(e, HdlPortItem):
                    # we are already added inputs of all components
                    continue
                else:
                    inputs = walkInputsForSpecificOutput(sig, e)
                for i in inputs:
                    if isinstance(i, RtlSignalBase) and i not in seen:
                        seen.add(i)
                        _toSearch.append(i)

            nv = sig._nop_val
            if isinstance(nv, RtlSignalBase):
                if nv not in seen:
                    seen.add(nv)
                    _toSearch.append(nv)
        toSearch = _toSearch

    seen.update([s for s, d in netlist.interfaces.items() if d == DIRECTION.IN])
    for c in netlist.subUnits:
        seen.update(sig._sig for sig in walkPhysInterfaces(c))

    for sig in netlist.signals:
        if sig in seen:
            continue

        for e in tuple(sig.drivers):
            # drivers of this signal are useless rm them
            if isinstance(e, Operator):
                removed_e = e
            elif isinstance(e, HdlPortItem):
                raise NotImplementedError(sig)
            else:
                removed_e = e._cut_off_drivers_of(sig)

            if removed_e is not None:
                # must not destroy before procesing inputs
                removed_e._destroy()

    netlist.signals = seen




from typing import Optional

from hdlConvertorAst.hdlAst._expr import HdlValueId
from hdlConvertorAst.hdlAst._structural import HdlModuleDef
from hdlConvertorAst.to.hwt.keywords import HWT_KEYWORDS
from hdlConvertorAst.translate.common.name_scope import LanguageKeyword, NameScope
from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import AllOps
from hwt.serializer.generic.to_hdl_ast import ToHdlAst
from hwt.serializer.hwt.ops import ToHdlAstHwt_ops
from hwt.serializer.hwt.types import ToHdlAstHwt_types
from hwt.serializer.hwt.value import ToHdlAstHwt_value
from hwt.serializer.simModel.serializer import ToHdlAstSimModel
from hwt.code import CodeBlock
from hdlConvertorAst.hdlAst import HdlStmBlock


class ToHdlAstHwt(ToHdlAstHwt_value, ToHdlAstHwt_ops,
                  ToHdlAstHwt_types, ToHdlAst):
    """
    Serializer which converts Hwt objects back to Hwt code
    for debugging purposes/code ports

    :ivar ~._valueWidthRequired: flag which tells if the values are required to have
        the width specified
    """
    _keywords_dict = {kw: LanguageKeyword() for kw in HWT_KEYWORDS}

    def __init__(self, name_scope: Optional[NameScope]=None):
        super(ToHdlAstHwt, self).__init__(name_scope=name_scope)
        self._valueWidthRequired = False
        self.currentUnit = None
        self.debug = False

    def has_to_be_process(self, proc):
        return True

    def can_pop_process_wrap(self, statements, hasToBeVhdlProcess):
        return False

    def _as_hdl_HdlModuleDef(self, new_m: HdlModuleDef) -> HdlModuleDef:
        return ToHdlAstSimModel._as_hdl_HdlModuleDef(self, new_m)

    def sensitivityListItem(self, item, anyIsEventDependnt):
        if isinstance(item, Operator):
            op = item.operator
            assert op in (AllOps.RISING_EDGE, AllOps.FALLING_EDGE), item
            assert not item.operands[0].hidden, item
            return self.as_hdl_Operator(item)
        else:
            return HdlValueId(item.name, obj=item)

    def as_hdl_CodeBlock(self, o: CodeBlock):
        res = HdlStmBlock()
        for _o in o.statements:
            res.body.append(self.as_hdl(_o))
        return res




from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import AllOps
from hwt.hdl.types.defs import BOOL
from hwt.hdl.value import HValue


class EventCapableVal():
    """
    Base class for event capable values
    """

    def _onFallingEdge(self):
        assert not isinstance(self, HValue), self
        return Operator.withRes(AllOps.FALLING_EDGE, [self], BOOL)

    def _onRisingEdge(self):
        assert not isinstance(self, HValue), self
        return Operator.withRes(AllOps.RISING_EDGE, [self], BOOL)



from typing import Union

from hdlConvertorAst.hdlAst._defs import HdlIdDef
from hdlConvertorAst.hdlAst._expr import HdlValueId, HdlOp, HdlOpType
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model.utils import hdl_call, \
    hdl_getattr
from hdlConvertorAst.translate.common.name_scope import LanguageKeyword
from hwt.code import Concat
from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import AllOps
from hwt.hdl.types.bitsVal import BitsVal
from hwt.hdl.types.defs import BIT
from hwt.hdl.types.enum import HEnum
from hwt.hdl.types.enumVal import HEnumVal
from hwt.hdl.value import HValue
from hwt.hdl.variables import SignalItem
from hwt.serializer.generic.ops import HWT_TO_HDLCONVEROTR_OPS
from hwt.serializer.generic.value import ToHdlAst_Value
from pyMathBitPrecise.array3t import Array3val
from pyMathBitPrecise.bits3t import Bits3val, Bits3t


zero, one = BIT.from_py(0), BIT.from_py(1)


class ToHdlAstSimModel_value(ToHdlAst_Value):
    Bits3val = HdlValueId("Bits3val", obj=Bits3val)
    Bits3t = HdlValueId("Bits3t", obj=Bits3t)
    SELF = HdlValueId("self", obj=LanguageKeyword())
    Array3val = HdlValueId("Array3val", obj=Array3val)
    SLICE = HdlValueId("slice", obj=slice)
    TRUE = HdlValueId("True", obj=True)
    FALSE = HdlValueId("False", obj=False)
    Bits3val = HdlValueId("Bits3val", obj=Bits3val)
    ABits3t = HdlValueId("Bits3t", obj=Bits3t)
    SELF_IO = hdl_getattr(HdlValueId("self", obj=LanguageKeyword()), "io")
    CONCAT = HdlValueId("Concat", obj=Concat)
    op_transl_dict = {
        **HWT_TO_HDLCONVEROTR_OPS,
        AllOps.INDEX: HdlOpType.INDEX,
    }
    _cast_ops = {
        AllOps.BitsAsSigned,
        AllOps.BitsAsUnsigned,
        AllOps.BitsAsVec,
    }

    def is_suitable_for_const_extract(self, val: HValue):
        return not isinstance(val._dtype, HEnum) or val.vld_mask == 0

    def as_hdl_SignalItem(self, si: Union[SignalItem, HdlIdDef],
                          declaration=False):
        if not declaration and not si.hidden:
            if si._const:
                return hdl_getattr(self.SELF, si.name)
            else:
                return hdl_getattr(hdl_getattr(self.SELF_IO, si.name), "val")
        else:
            return super(ToHdlAstSimModel_value, self).as_hdl_SignalItem(
                si, declaration=declaration)

    def as_hdl_BitsVal(self, val: BitsVal):
        dtype = val._dtype
        as_hdl_int = self.as_hdl_int
        t = hdl_call(self.Bits3t, [
            as_hdl_int(dtype.bit_length()), as_hdl_int(int(bool(dtype.signed)))])
        return hdl_call(self.Bits3val, [t,
                                        as_hdl_int(val.val),
                                        as_hdl_int(val.vld_mask)])

    def as_hdl_DictVal(self, val):
        return {
            self.as_hdl_int(int(k)): self.as_hdl_Value(v)
            for k, v in val.items()
        }

    def as_hdl_HArrayVal(self, val):
        return hdl_call(self.Array3val, [
            self.as_hdl_HdlType(val._dtype),
            self.as_hdl_DictVal(val.val),
            self.as_hdl_int(val.vld_mask)
        ])

    def as_hdl_HSliceVal(self, val):
        args = (
            val.val.start,
            val.val.stop,
            val.val.step
        )
        return hdl_call(self.SLICE, [self.as_hdl_int(int(a)) for a in args])

    def as_hdl_HEnumVal(self, val: HEnumVal):
        t_name = self.name_scope.get_object_name(val._dtype)
        if val.vld_mask:
            name = self.name_scope.get_object_name(val)
            return hdl_getattr(hdl_getattr(self.SELF, t_name), name)
        else:
            return hdl_call(hdl_getattr(hdl_getattr(self.SELF, t_name), "from_py"),
                            [None, ])

    def as_hdl_Operator(self, op: Operator):
        ops = op.operands
        o = op.operator

        if o == AllOps.EQ:
            op0 = self.as_hdl_Value(ops[0])
            op1 = self.as_hdl_Value(ops[1])
            return hdl_call(hdl_getattr(op0, "_eq"), [op1, ])
        elif o == AllOps.TERNARY:
            if ops[1] == one and ops[2] == zero:
                # ignore redundant x ? 1 : 0
                return self.as_hdl_cond(ops[0], True)
            else:
                op0 = self.as_hdl_cond(ops[0], True)
                op1 = self.as_hdl_Value(ops[1])
                op2 = self.as_hdl_Value(ops[2])
                return hdl_call(hdl_getattr(op0, "_ternary"), [op1, op2])
        elif o == AllOps.RISING_EDGE or o == AllOps.FALLING_EDGE:
            if o == AllOps.RISING_EDGE:
                fn = "_onRisingEdge"
            else:
                fn = "_onFallingEdge"
            op0 = self.as_hdl_Value(ops[0])
            # pop .val
            op0 = op0.ops[0]
            return hdl_call(hdl_getattr(op0, fn), [])
        elif o in self._cast_ops:
            op0, = ops
            do_cast = bool(op0._dtype.signed) != bool(op.result._dtype.signed)

            op_hdl = self.as_hdl_Value(op0)
            if do_cast:
                if bool(op.result._dtype.signed):
                    sign = self.TRUE
                else:
                    sign = self.FALSE
                return hdl_call(hdl_getattr(op_hdl, "cast_sign"), [sign, ])
            else:
                return op_hdl
        elif o == AllOps.CONCAT:
            return hdl_call(hdl_getattr(self.as_hdl_Value(ops[0]), "_concat"),
                            [self.as_hdl_Value(ops[1]), ])
        elif o == AllOps.EQ:
            return hdl_call(hdl_getattr(self.as_hdl_Value(ops[0]), "_eq"),
                            [self.as_hdl_Value(ops[1]), ])
        else:
            o = self.op_transl_dict[o]
            return HdlOp(o, [self.as_hdl_Value(o2)
                               for o2 in ops])
